"""Module containing different I/O functions for biomarker data (saliva, dried blood spots, IL-6)."""

from collections.abc import Sequence
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd

from biopsykit.io.io import _apply_index_cols
from biopsykit.utils._datatype_validation_helper import _assert_file_extension, _assert_has_columns
from biopsykit.utils._types_internal import path_t
from biopsykit.utils.dtypes import (
    BiomarkerRawDataFrame,
    SalivaRawDataFrame,
    SubjectConditionDataFrame,
    _BiomarkerRawDataFrame,
    _SalivaRawDataFrame,
    _SubjectConditionDataFrame,
    is_biomarker_raw_dataframe,
    is_saliva_raw_dataframe,
    is_subject_condition_dataframe,
)

__all__ = ["load_biomarker_results", "load_saliva_plate", "load_saliva_wide_format", "save_saliva"]

_DATA_COL_NAMES = {
    "cortisol": "cortisol (nmol/l)",
    "amylase": "amylase (U/ml)",
    "crp": "CRP (µg/ml)",
}


def load_saliva_plate(
    file_path: path_t,
    saliva_type: str,
    sample_id_col: Optional[str] = None,
    data_col: Optional[str] = None,
    id_col_names: Optional[Sequence[str]] = None,
    regex_str: Optional[str] = None,
    sample_times: Optional[Sequence[int]] = None,
    condition_list: Optional[Union[Sequence, dict[str, Sequence], pd.Index]] = None,
    **kwargs,
) -> SalivaRawDataFrame:
    r"""Read saliva from an Excel sheet in 'plate' format. Wraps load_biomarker_results() for compatibilty.

    This function automatically extracts identifier like subject, day and sample IDs from the saliva sample names.
    To extract them, a regular expression string can be passed via ``regex_str``.

    Here are some examples on how sample identifiers might look like and what the corresponding
    ``regex_str`` would output:

        * "Vp01 S1"
          => ``r"(Vp\d+) (S\d)"`` (this is the default pattern, you can also just set ``regex_str`` to ``None``)
          => data ``[Vp01, S1]`` in two columns: ``subject``, ``sample``
          (unless column names are explicitly specified in ``data_col_names``)
        * "Vp01 T1 S1" ... "Vp01 T1 S5" (only *numeric* characters in day/sample)
          => ``r"(Vp\d+) (T\d) (S\d)"``
          => three columns: ``subject``, ``sample`` with data ``[Vp01, T1, S1]``
          (unless column names are explicitly specified in ``data_col_names``)
        * "Vp01 T1 S1" ... "Vp01 T1 SA" (also *letter* characters in day/sample)
          => ``r"(Vp\d+) (T\w) (S\w)"``
          => three columns: ``subject``, ``sample`` with data ``[Vp01, T1, S1]``
          (unless column names are explicitly specified in ``data_col_names``)

    If you **don't** want to extract the 'S' or 'T' prefixes in saliva or day IDs, respectively,
    you have to move it **out** of the capture group in the ``regex_str`` (round brackets), like this:
    ``(S\d)`` (would give ``S1``, ``S2``, ...)
    => ``S(\d)`` (would give ``1``, ``2``, ...)


    Parameters
    ----------
    file_path: :class:`~pathlib.Path` or str
        path to the Excel sheet in 'plate' format containing saliva data
    saliva_type: str
        saliva type to load from file
    sample_id_col: str, optional
        column name of the Excel sheet containing the sample ID. Default: "sample ID"
    data_col: str, optional
        column name of the Excel sheet containing saliva data to be analyzed.
        Default: Select default column name based on ``biomarker_type``, e.g. ``cortisol`` => ``cortisol (nmol/l)``
    id_col_names: list of str, optional
        names of the extracted ID column names. ``None`` to use the default column names (['subject', 'day', 'sample'])
    regex_str: str, optional
        regular expression to extract subject ID, day ID and sample ID from the sample identifier.
        ``None`` to use default regex string (``r"(Vp\d+) (S\d)"``)
    sample_times: list of int, optional
        times at which saliva samples were collected
    condition_list: 1d-array, optional
        list of conditions which subjects were assigned to
    **kwargs
        Additional parameters that are passed to :func:`pandas.read_excel`

    Returns
    -------
    data : :class:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no Excel file (.xls or .xlsx)
    ValueError
        if any saliva sample can not be converted into a float (e.g. because there was text in one of the columns)
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if imported data can not be parsed to a SalivaRawDataFrame

    """
    if regex_str is None:
        regex_str = r"(Vp\d+) (S\d)"

    return _SalivaRawDataFrame(
        load_biomarker_results(
            file_path,
            biomarker_type=saliva_type,
            sample_id_col=sample_id_col,
            data_col=data_col,
            id_col_names=id_col_names,
            regex_str=regex_str,
            sample_times=sample_times,
            condition_list=condition_list,
            **kwargs,
        )
    )


def save_saliva(
    file_path: path_t,
    data: SalivaRawDataFrame,
    saliva_type: Optional[str] = "cortisol",
    as_wide_format: Optional[bool] = False,
):
    """Save saliva data to csv file.

    Parameters
    ----------
    file_path: :class:`~pathlib.Path` or str
        file path to export. Must be a csv or an Excel file
    data : :class:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str
        type of saliva data in the dataframe
    as_wide_format : bool, optional
        ``True`` to save data in wide format (and flatten all index levels), ``False`` to save data in long-format.
        Default: ``False``

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a SalivaRawDataFrame
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if ``file_path`` is not a csv or Excel file

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, [".csv", ".xls", ".xlsx"])

    is_saliva_raw_dataframe(data, saliva_type)
    data = data[saliva_type]
    if as_wide_format:
        levels = list(data.index.names)
        levels.remove("subject")
        data = data.unstack(level=levels)
        data.columns = ["_".join(col) for col in data.columns]

    if file_path.suffix in [".csv"]:
        data.to_csv(file_path)
    else:
        writer = pd.ExcelWriter(file_path, engine="xlsxwriter")  # pylint:disable=abstract-class-instantiated
        data.to_excel(writer)
        writer.close()


def load_saliva_wide_format(
    file_path: path_t,
    saliva_type: str,
    subject_col: Optional[str] = None,
    condition_col: Optional[str] = None,
    additional_index_cols: Optional[Union[str, Sequence[str]]] = None,
    sample_times: Optional[Sequence[int]] = None,
    **kwargs,
) -> SalivaRawDataFrame:
    """Load saliva data that is in wide-format from csv file.

    It will return a `SalivaRawDataFrame`, a long-format dataframe that complies with BioPsyKit's naming convention,
    i.e., the subject ID index will be named ``subject``, the sample index will be names ``sample``,
    and the value column will be named after the saliva biomarker type.

    Parameters
    ----------
    file_path: :class:`~pathlib.Path` or str
        path to file
    saliva_type: str
        saliva type to load from file. Example: ``cortisol``
    subject_col: str, optional
        name of column containing subject IDs or ``None`` to use the default column name ``subject``.
        According to BioPsyKit's convention, the subject ID column is expected to have the name ``subject``.
        If the subject ID column in the file has another name, the column will be renamed in the dataframe
        returned by this function. Default: ``None``
    condition_col : str, optional
        name of the column containing condition assignments or ``None`` if no conditions are present.
        According to BioPsyKit's convention, the condition column is expected to have the name ``condition``.
        If the condition column in the file has another name, the column will be renamed in the dataframe
        returned by this function. Default: ``None``
    additional_index_cols : str or list of str, optional
        additional index levels to be added to the dataframe, e.g., "day" index. Can either be a string or a list
        strings to indicate column name(s) that should be used as index level(s),
        or ``None`` for no additional index levels. Default: ``None``
    sample_times: list of int, optional
        times at which saliva samples were collected or ``None`` if no sample times should be specified.
        Default: ``None``
    **kwargs
        Additional parameters that are passed to :func:`pandas.read_csv` or :func:`pandas.read_excel`

    Returns
    -------
    data : :class:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no csv or Excel file

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, [".csv", ".xls", ".xlsx"])
    data = _read_dataframe(file_path, **kwargs)

    if subject_col is None:
        subject_col = "subject"

    _assert_has_columns(data, [[subject_col]])

    if subject_col != "subject":
        # rename column
        data = data.rename(columns={subject_col: "subject"})
        subject_col = "subject"

    index_cols = [subject_col]

    data, condition_col = _get_condition_col(data, condition_col)

    index_cols = _get_index_cols(condition_col, index_cols, additional_index_cols)
    data = _apply_index_cols(data, index_cols=index_cols)

    num_subjects = len(data)
    data.columns = pd.MultiIndex.from_product([[saliva_type], data.columns], names=[None, "sample"])
    data = data.stack(future_stack=True)

    _check_num_samples(len(data), num_subjects)

    if sample_times is not None:
        _check_sample_times(len(data), num_subjects, sample_times)
        data["time"] = np.array(sample_times * num_subjects)

    is_saliva_raw_dataframe(data, saliva_type)

    return _SalivaRawDataFrame(data)


def load_biomarker_results(
    file_path: path_t,
    biomarker_type: Optional[str] = None,
    sample_id_col: Optional[str] = None,
    data_col: Optional[str] = None,
    id_col_names: Optional[Sequence[str]] = None,
    regex_str: Optional[str] = None,
    sample_times: Optional[Sequence[int]] = None,
    condition_list: Optional[Union[Sequence, dict[str, Sequence], pd.Index]] = None,
    **kwargs,
) -> BiomarkerRawDataFrame:
    r"""Load biomarker results from Excel file.

    Parameters
    ----------
    file_path: :class:`~pathlib.Path` or str
        path to file
    dbs_type: str, optional
        biomarker type to load from file. Example: ``crp``
    sample_id_col: str, optional
        name of column containing sample IDs or ``None`` to use the default column name ``sample ID``.
    data_col: str, optional
        name of column containing biomarker data or ``None`` to use the default column name.
    id_col_names: list of str, optional
        names of the extracted ID column names. ``None`` to use the default column names (['subject', 'day', 'sample'])
    regex_str: str, optional
        regular expression to extract subject, day, and sample ID from sample ID column.
        ``None`` to use the default regular expression ``r"(VP\\d+)-(T\\w)-(B\\w)"``.
    sample_times: list of int, optional
        times at which samples were collected or ``None`` if no sample times should be specified.
        Default: ``None``
    condition_list: list of str, dict of str to list of str, or :class:`~pandas.Index`, optional
        list of condition names or dictionary of condition names to list of condition assignments or
        :class:`~pandas.Index` of condition names or ``None`` if no conditions are present.
        Default: ``None``
    **kwargs
        Additional parameters that are passed to :func:`pandas.read_csv` or :func:`pandas.read_excel`

    Returns
    -------
    data : :class:`~biopsykit.utils.dtypes.BiomarkerRawDataFrame`
        biomarker data in `BiomarkerRawDataFrame` format

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no Excel file
    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, (".xls", ".xlsx"))

    # TODO add remove_nan option (all or any)
    if regex_str is None:
        regex_str = r"(VP\d+)-(T\w)-(B\w)"

    if sample_id_col is None:
        sample_id_col = "sample ID"

    if data_col is None:
        data_col = _DATA_COL_NAMES[biomarker_type]

    df_biomarker = pd.read_excel(file_path, skiprows=2, usecols=[sample_id_col, data_col], **kwargs)
    cols = df_biomarker[sample_id_col].str.extract(regex_str)
    id_col_names = _get_id_columns(id_col_names, cols)

    df_biomarker[id_col_names] = cols

    df_biomarker = df_biomarker.drop(columns=[sample_id_col], errors="ignore")
    df_biomarker = df_biomarker.rename(columns={data_col: biomarker_type})
    df_biomarker = df_biomarker.set_index(id_col_names)

    if condition_list is not None:
        df_biomarker = _apply_condition_list(df_biomarker, condition_list)

    num_subjects = len(df_biomarker.index.get_level_values("subject").unique())

    _check_num_samples(len(df_biomarker), num_subjects)

    if sample_times:
        _check_sample_times(len(df_biomarker), num_subjects, sample_times)
        df_biomarker["time"] = np.array(sample_times * num_subjects)

    try:
        df_biomarker[biomarker_type] = df_biomarker[biomarker_type].astype(float)
    except ValueError as e:
        raise ValueError(
            f"""Error converting all biomarker values into numbers: '{e}'
            Please check your biomarker values whether there is any text etc. in the column '{data_col}'
            and delete the values or replace them by NaN!"""
        ) from e

    is_biomarker_raw_dataframe(df_biomarker, biomarker_type=biomarker_type)

    return _BiomarkerRawDataFrame(df_biomarker)


def _get_index_cols(condition_col: str, index_cols: Sequence[str], additional_index_cols: Sequence[str]):
    if condition_col is not None:
        index_cols = [condition_col, *index_cols]

    if additional_index_cols is None:
        additional_index_cols = []
    if isinstance(additional_index_cols, str):
        additional_index_cols = [additional_index_cols]

    return index_cols + additional_index_cols


def _read_dataframe(file_path: Path, **kwargs):
    if file_path.suffix in [".csv"]:
        return pd.read_csv(file_path, **kwargs)
    return pd.read_excel(file_path, **kwargs)


def _check_num_samples(num_samples: int, num_subjects: int):
    """Check that number of imported samples is the same for all subjects.

    Parameters
    ----------
    num_samples : int
        total number of saliva samples in the current dataframe
    num_subjects : int
        total number of subjects in the current dataframe

    Raises
    ------
    ValueError
        if number of samples is not equal for all subjects

    """
    if num_samples % num_subjects != 0:
        raise ValueError(
            f"Error during import: Number of samples not equal for all subjects! "
            f"Got {num_samples} samples for {num_subjects} subjects."
        )


def _check_sample_times(num_samples: int, num_subjects: int, sample_times: Sequence[int]):
    """Check that sample times have the correct number of samples and are monotonously increasing.

    Parameters
    ----------
    num_samples : int
        total number of saliva samples in the current dataframe
    num_subjects : int
        total number of subjects in the current dataframe
    sample_times : array-like
        list of sample times

    Raises
    ------
    ValueError
        if values in ``sample_times`` are not monotonously increasing or
        if number of saliva times does not match the number of saliva samples per subject

    """
    if np.any(np.diff(sample_times) <= 0):
        raise ValueError("`saliva_times` must be increasing!")
    if (len(sample_times) * num_subjects) != num_samples:
        raise ValueError(
            f"Length of `saliva_times` does not match the number of saliva samples! "
            f"Expected: {int(num_samples / num_subjects)}, got: {len(sample_times)}"
        )


def _parse_condition_list(
    data: pd.DataFrame, condition_list: Union[Sequence, dict[str, Sequence], pd.Index]
) -> SubjectConditionDataFrame:
    if isinstance(condition_list, (list, np.ndarray)):
        # Add Condition as new index level
        condition_list = pd.DataFrame(
            condition_list,
            columns=["condition"],
            index=data.index.get_level_values("subject").unique(),
        )
    elif isinstance(condition_list, dict):
        condition_list = [(subject, cond) for cond in condition_list for subject in condition_list[cond]]
        condition_list = pd.DataFrame(condition_list, columns=["subject", "condition"])
        condition_list = condition_list.set_index("subject")
    elif isinstance(condition_list, pd.DataFrame):
        condition_list = condition_list.reset_index().set_index("subject")

    is_subject_condition_dataframe(condition_list)
    return _SubjectConditionDataFrame(condition_list)


def _apply_condition_list(
    data: pd.DataFrame,
    condition_list: Optional[Union[Sequence, dict[str, Sequence], pd.Index]] = None,
):
    condition_list = _parse_condition_list(data, condition_list)

    data = (
        data.join(condition_list).set_index("condition", append=True).reorder_levels(["condition", "subject", "sample"])
    )
    return data


def _get_id_columns(id_col_names: Sequence[str], extracted_cols: pd.DataFrame):
    if id_col_names is None:
        id_col_names = ["subject", "sample"]
        if len(extracted_cols.columns) == 3:
            id_col_names = ["subject", "day", "sample"]
    elif len(id_col_names) != len(extracted_cols.columns):
        raise ValueError(
            f"Number of 'id_col_names' must match length of extracted index columns! "
            f"Expected {len(extracted_cols)}, got {len(id_col_names)}."
        )

    return id_col_names


def _get_condition_col(data: pd.DataFrame, condition_col: str) -> tuple[pd.DataFrame, str]:
    if condition_col is None:
        if "condition" in data.columns:
            condition_col = "condition"
    else:
        _assert_has_columns(data, [[condition_col]])
        if condition_col != "condition":
            # rename column
            data = data.rename(columns={condition_col: "condition"})
            condition_col = "condition"
    return data, condition_col



"""A couple of helper functions that ease the use of the typical biopsykit data formats."""

from typing import Any, Optional, Union

import numpy as np
import pandas as pd

from biopsykit.utils._datatype_validation_helper import (
    _assert_has_column_multiindex,
    _assert_has_column_prefix,
    _assert_has_columns,
    _assert_has_columns_any_level,
    _assert_has_index_levels,
    _assert_has_multiindex,
    _assert_is_dtype,
    _assert_sample_columns_int,
)
from biopsykit.utils.exceptions import ValidationError

__all__ = [
    "ECG_RESULT_DATAFRAME_COLUMNS",
    "HEART_RATE_DATAFRAME_COLUMNS",
    "R_PEAK_DATAFRAME_COLUMNS",
    "Acc1dDataFrame",
    "Acc3dDataFrame",
    "BPointDataFrame",
    "BiomarkerRawDataFrame",
    "CPointDataFrame",
    "CodebookDataFrame",
    "EcgRawDataFrame",
    "EcgResultDataFrame",
    "Gyr1dDataFrame",
    "Gyr3dDataFrame",
    "HeartRateDataFrame",
    "HeartRatePhaseDict",
    "HeartRateStudyDataDict",
    "HeartRateSubjectDataDict",
    "HeartbeatSegmentationDataFrame",
    "IcgRawDataFrame",
    "ImuDataFrame",
    "MeanSeDataFrame",
    "MergedStudyDataDict",
    "PhaseDict",
    "QPeakDataFrame",
    "RPeakDataFrame",
    "SalivaFeatureDataFrame",
    "SalivaMeanSeDataFrame",
    "SalivaRawDataFrame",
    "SleepEndpointDataFrame",
    "SleepEndpointDict",
    "SleepWakeDataFrame",
    "StudyDataDict",
    "SubjectConditionDataFrame",
    "SubjectConditionDict",
    "SubjectDataDict",
    "is_acc1d_dataframe",
    "is_acc3d_dataframe",
    "is_b_point_dataframe",
    "is_biomarker_raw_dataframe",
    "is_c_point_dataframe",
    "is_codebook_dataframe",
    "is_ecg_raw_dataframe",
    "is_ecg_result_dataframe",
    "is_gyr1d_dataframe",
    "is_gyr3d_dataframe",
    "is_heart_rate_dataframe",
    "is_heartbeat_segmentation_dataframe",
    "is_hr_phase_dict",
    "is_hr_subject_data_dict",
    "is_icg_raw_dataframe",
    "is_imu_dataframe",
    "is_mean_se_dataframe",
    "is_merged_study_data_dict",
    "is_pep_result_dataframe",
    "is_phase_dict",
    "is_q_peak_dataframe",
    "is_r_peak_dataframe",
    "is_saliva_feature_dataframe",
    "is_saliva_mean_se_dataframe",
    "is_saliva_raw_dataframe",
    "is_sleep_endpoint_dataframe",
    "is_sleep_endpoint_dict",
    "is_sleep_wake_dataframe",
    "is_study_data_dict",
    "is_subject_condition_dataframe",
    "is_subject_condition_dict",
    "is_subject_data_dict",
]

ECG_RESULT_DATAFRAME_COLUMNS = ["ECG_Raw", "ECG_Clean", "ECG_Quality", "ECG_R_Peaks", "R_Peak_Outlier"]
HEART_RATE_DATAFRAME_COLUMNS = ["Heart_Rate"]
R_PEAK_DATAFRAME_COLUMNS = ["R_Peak_Quality", "R_Peak_Idx", "RR_Interval", "R_Peak_Outlier"]

PEP_RESULT_DATAFRAME_COLUMNS = [
    "heartbeat_start_sample",
    "heartbeat_end_sample",
    "r_peak_sample",
    "q_peak_sample",
    "b_point_sample",
    "pep_sample",
    "pep_ms",
]

# these subclasses of pd.DataFrame are needed to be added to the type aliases because otherwise, autosphinx does not
# add the docstring to the documentation of the type aliases. Additionally, they can be used internally to highlight
# which alias types are expected at which position


class _SubjectConditionDataFrame(pd.DataFrame):
    pass


class _CodebookDataFrame(pd.DataFrame):
    pass


class _MeanSeDataFrame(pd.DataFrame):
    pass


class _BiomarkerRawDataFrame(pd.DataFrame):
    pass


class _SalivaRawDataFrame(pd.DataFrame):
    pass


class _SalivaFeatureDataFrame(pd.DataFrame):
    pass


class _SalivaMeanSeDataFrame(pd.DataFrame):
    pass


class _IcgRawDataFrame(pd.DataFrame):
    pass


class _EcgRawDataFrame(pd.DataFrame):
    pass


class _EcgResultDataFrame(pd.DataFrame):
    pass


class _HeartRateDataFrame(pd.DataFrame):
    pass


class _RPeakDataFrame(pd.DataFrame):
    pass


class _Acc1dDataFrame(pd.DataFrame):
    pass


class _Acc3dDataFrame(pd.DataFrame):
    pass


class _Gyr1dDataFrame(pd.DataFrame):
    pass


class _Gyr3dDataFrame(pd.DataFrame):
    pass


class _ImuDataFrame(pd.DataFrame):
    pass


class _SleepWakeDataFrame(pd.DataFrame):
    pass


class _SleepEndpointDataFrame(pd.DataFrame):
    pass


class _HeartbeatSegmentationDataFrame(pd.DataFrame):
    pass


class _QPeakDataFrame(pd.DataFrame):
    pass


class _BPointDataFrame(pd.DataFrame):
    pass


class _CPointDataFrame(pd.DataFrame):
    pass


class _PepResultDataFrame(pd.DataFrame):
    pass


SubjectConditionDataFrame = Union[_SubjectConditionDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing subject IDs and condition assignment in a standardized format.

A ``SubjectConditionDataFrame`` has an index with subject IDs named ``subject`` and a column with the condition
assignment named ``condition``.
"""

SubjectConditionDict = dict[str, np.ndarray]
"""Dictionary containing subject IDs and condition assignment in a standardized format.

A ``SubjectConditionDict`` contains conditions as dictionary keys and a collection of subject IDs
(list, numpy array, pandas Index) as dictionary values.
"""

CodebookDataFrame = Union[_CodebookDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` representing a codebook which encodes numerical and categorical values
in a standardized format.

A ``CodebookDataFrame`` has an index level named ``variable``. The column names are the numerical values (0, 1, ...),
the dataframe entries then represent the mapping of numerical value to categorical value for the variable.
"""

MeanSeDataFrame = Union[_MeanSeDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing mean and standard error of time-series data in a standardized format.

The resulting dataframe must at least the two columns ``mean`` and ``se``. It can have additional index levels,
such as ``phase``, ``subphase`` or ``condition``.
"""

BiomarkerRawDataFrame = Union[_BiomarkerRawDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing raw biomarker data in a standardized format.

Data needs to be in long-format and **must** have a :class:`pandas.MultiIndex` with index level names:

* ``subject``: subject ID; can be number or string
* ``sample``: saliva sample ID; can be number or string

Additionally, the following index levels can be added to identify saliva values, such as:

* ``condition``: subject condition during the study (e.g., "Control" vs. "Condition")
* ``day``: day ID, if saliva samples were collected over multiple days
* ``night``: night ID, if saliva samples were collected over multiple night
* ...

"""

SalivaRawDataFrame = Union[_BiomarkerRawDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing raw saliva data in a standardized format.

Data needs to be in long-format and **must** have a :class:`pandas.MultiIndex` with index level names:

* ``subject``: subject ID; can be number or string
* ``sample``: saliva sample ID; can be number or string

Additionally, the following index levels can be added to identify saliva values, such as:

* ``condition``: subject condition during the study (e.g., "Control" vs. "Condition")
* ``day``: day ID, if saliva samples were collected over multiple days
* ``night``: night ID, if saliva samples were collected over multiple night
* ...

"""

SalivaFeatureDataFrame = Union[_SalivaFeatureDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing feature computed from saliva data in a standardized format.

The resulting dataframe must at least have a ``subject`` index level and all column names need to begin with
the saliva marker type (e.g. "cortisol"), followed by the feature name, separated by underscore '_'
Additionally, the name of the column index needs to be `saliva_feature`.
"""

SalivaMeanSeDataFrame = Union[_SalivaMeanSeDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing mean and standard error of saliva samples in a standardized format.

The resulting dataframe must at least have a ``sample`` index level and the two columns ``mean`` and ``se``.
It can have additional index levels, such as ``condition`` or ``time``.
"""

SleepEndpointDict = dict[str, Any]
"""Dictionary containing sleep endpoints in a standardized format.

The dict entries represent the sleep endpoints and should follow a standardized naming convention,
regardless of the origin (IMU sensor, sleep mattress, psg, etc.).

Required are the entries:

* ``sleep_onset``: Sleep Onset, i.e., time of falling asleep, in absolute time
* ``wake_onset``: Wake Onset, i.e., time of awakening, in absolute time
* ``total_sleep_duration``: Total sleep duration, i.e., time between Sleep Onset and Wake Onset, in minutes

The following entries are common, but not required:

* ``total_duration``: Total recording time, in minutes
* ``net_sleep_duration``: Net duration spent sleeping, in minutes
* ``bed_interval_start``: Bed Interval Start, i.e, time when participant went to bed, in absolute time
* ``bed_interval_end``: Bed Interval End, i.e, time when participant left bed, in absolute time
* ``sleep_efficiency``: Sleep Efficiency, defined as the ratio between net sleep duration and sleep duration
  in percent
* ``sleep_onset_latency``: Sleep Onset Latency, i.e., time in bed needed to fall asleep, in minutes
* ``getup_latency``: Get Up Latency, i.e., time in bed after awakening until getting up, in minutes
* ``wake_after_sleep_onset``: Wake After Sleep Onset (WASO), i.e., total time awake after falling asleep, in minutes
* ``sleep_bouts``: List with start and end times of sleep bouts
* ``wake_bouts``: List with start and end times of wake bouts
* ``number_wake_bouts``: Total number of wake bouts

The following entries are, for instance, further possible:

* ``total_time_light_sleep``: Total time of light sleep, in minutes
* ``total_time_deep_sleep``: Total time of deep sleep, in minutes
* ``total_time_rem_sleep``: Total time of REM sleep, in minutes
* ``total_time_awake``: Total time of being awake, in minutes
* ``count_snoring_episodes``: Total number of snoring episodes
* ``total_time_snoring``: Total time of snoring, in minutes
* ``heart_rate_avg``: Average heart rate during recording, in bpm
* ``heart_rate_min``: Minimum heart rate during recording, in bpm
* ``heart_rate_max``: Maximum heart rate during recording, in bpm

"""

SleepEndpointDataFrame = Union[_SleepEndpointDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing sleep endpoints in a standardized format.

The resulting dataframe must at least have a ``date`` index level,
and, optionally, further index levels like ``night``.

The columns defining the sleep endpoints should follow a standardized naming convention, regardless of the origin
(IMU sensor, sleep mattress, psg, etc.).

Required are the columns:

* ``sleep_onset``: Sleep Onset, i.e., time of falling asleep, in absolute time
* ``wake_onset``: Wake Onset, i.e., time of awakening, in absolute time
* ``total_sleep_duration``: Total sleep duration, i.e., time between Sleep Onset and Wake Onset, in minutes

The following columns are common, but not required:

* ``total_duration``: Total recording time, in minutes
* ``net_sleep_duration``: Net duration spent sleeping, in minutes
* ``bed_interval_start``: Bed Interval Start, i.e, time when participant went to bed, in absolute time
* ``bed_interval_end``: Bed Interval End, i.e, time when participant left bed, in absolute time
* ``sleep_efficiency``: Sleep Efficiency, defined as the ratio between net sleep duration and sleep duration
  in percent
* ``sleep_onset_latency``: Sleep Onset Latency, i.e., time in bed needed to fall asleep, in minutes
* ``getup_latency``: Get Up Latency, i.e., time in bed after awakening until getting up, in minutes
* ``wake_after_sleep_onset``: Wake After Sleep Onset (WASO), i.e., total time awake after falling asleep, in minutes
* ``number_wake_bouts``: Total number of wake bouts

The following columns are further possible:

* ``total_time_light_sleep``: Total time of light sleep, in minutes
* ``total_time_deep_sleep``: Total time of deep sleep, in minutes
* ``total_time_rem_sleep``: Total time of REM sleep, in minutes
* ``total_time_awake``: Total time of being awake, in minutes
* ``count_snoring_episodes``: Total number of snoring episodes
* ``total_time_snoring``: Total time of snoring, in minutes
* ``heart_rate_avg``: Average heart rate during recording, in bpm
* ``heart_rate_min``: Minimum heart rate during recording, in bpm
* ``heart_rate_max``: Maximum heart rate during recording, in bpm

"""

IcgRawDataFrame = Union[_IcgRawDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing raw ICG data of `one` subject.

The dataframe is expected to have one of the following columns:

* ``icg``: Raw ICG signal
* ``icg_der``: Derivative of the ICG signal

"""


EcgRawDataFrame = Union[_EcgRawDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing raw ECG data of `one` subject.

The dataframe is expected to have the following columns:

* ``ecg``: Raw ECG signal

"""

EcgResultDataFrame = Union[_EcgResultDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing processed ECG data of `one` subject.

The dataframe is expected to have the following columns:

* ``ECG_Raw``: Raw ECG signal
* ``ECG_Clean``: Cleaned (filtered) ECG signal
* ``ECG_Quality``: ECG signal quality indicator in the range of [0, 1]
* ``ECG_R_Peaks``: 1.0 where R peak was detected in the ECG signal, 0.0 else
* ``R_Peak_Outlier``: 1.0 when a detected R peak was classified as outlier, 0.0 else
* ``Heart_Rate``: Computed Heart rate interpolated to signal length

"""

HeartRateDataFrame = Union[_HeartRateDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing heart rate time series data of `one` subject.

The dataframe is expected to have the following columns:

* ``Heart_Rate``: Heart rate data. Can either be instantaneous heart rate or resampled heart rate

"""

RPeakDataFrame = Union[_RPeakDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing R-peak locations of `one` subject extracted from ECG data.

The dataframe is expected to have the following columns:

* ``R_Peak_Quality``: Signal quality indicator (of the raw ECG signal) in the range of [0, 1]
* ``R_Peak_Idx``: Array index of detected R peak in the raw ECG signal
* ``RR_Interval``: Interval between the current and the successive R peak in seconds
* ``R_Peak_Outlier``: 1.0 when a detected R peak was classified as outlier, 0.0 else

"""

Acc1dDataFrame = Union[_Acc1dDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 1-d acceleration data.

The dataframe is expected to have one of the following column sets:

* ["acc"]: one level column index
* ["acc_norm"]: one level column index

"""

Acc3dDataFrame = Union[_Acc3dDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 3-d acceleration data.

The dataframe is expected to have one of the following column sets:

* ["acc_x", "acc_y", "acc_z"]: one level column index
* [("acc", "x"), ("acc", "y"), ("acc", "z")]: two-level column index, first level specifying the channel
  (acceleration), second level specifying the axes

"""

Gyr1dDataFrame = Union[_Gyr1dDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 1-d gyroscope data.

The dataframe is expected to have one of the following column sets:

* ["gyr"]: one level column index
* ["gyr_norm"]: one level column index

"""

Gyr3dDataFrame = Union[_Gyr3dDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 3-d gyroscope data.

The dataframe is expected to have one of the following column sets:

* ["gyr_x", "gyr_y", "gyr_z"]: one level column index
* [("gyr", "x"), ("gyr", "y"), ("gyr", "z")]: two-level column index, first level specifying the channel
  (gyroscope), second level specifying the axes

"""

ImuDataFrame = Union[_ImuDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 6-d inertial measurement (IMU) (acceleration and gyroscope) data.

Hence, an ``ImuDataFrame`` must both be a ``AccDataFrame`` **and** a ``GyrDataFrame``.

The dataframe is expected to have one of the following column sets:

* ["acc_x", "acc_y", "acc_z", "gyr_x", "gyr_y", "gyr_z"]: one level column index
* [("acc", "x"), ("acc", "y"), ("acc", "z"), ("gyr", "x"), ("gyr", "y"), ("gyr", "z")]:
  two-level column index, first level specifying the channel (acceleration and gyroscope),
  second level specifying the axes

"""

SleepWakeDataFrame = Union[_SleepWakeDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing sleep/wake predictions.

The dataframe is expected to have at least the following column(s):

* ["sleep_wake"]: sleep/wake predictions where 1 indicates sleep and 0 indicates wake

"""

HeartbeatSegmentationDataFrame = Union[_HeartbeatSegmentationDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing results of heartbeat segmentation.

The dataframe is expected to have *at least* the following columns:

* ``start_sample``: Start sample of segmented heartbeat
* ``end_sample``: End sample of segmented heartbeat
* ``r_peak_sample``: R-peak sample of segmented heartbeat

"""

QPeakDataFrame = Union[_QPeakDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing Q-peak locations extracted from ECG data.

The dataframe is expected to have *at least* the following columns:

* ``q_peak_sample``: The sample index of the Q-peak in the ECG signal

Optionally, the dataframe can contain additional columns, such as:

* ``nan_reason``: Reason why the Q-peak was set to NaN (e.g., "r_peak_nan", "no_zero_crossing")

"""

BPointDataFrame = Union[_BPointDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing B-Point locations extracted from ICG data.

The dataframe is expected to have *at least* the following columns:

* ``b_point_sample``: The sample index of the B-point in the ICG signal

Optionally, the dataframe can contain additional columns, such as:

* ``nan_reason``: Reason why the B-point was set to NaN (e.g., "c_point_nan", "no_zero_crossing")

"""

CPointDataFrame = Union[_CPointDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing C-Point locations extracted from ICG data.

The dataframe is expected to have *at least* the following columns:

* ``c_point_sample``: The sample index of the C-point in the ICG signal

Optionally, the dataframe can contain additional columns, such as:

* ``nan_reason``: Reason why the C-point was set to NaN (e.g., "no_local_maxima")

"""

PepResultDataFrame = Union[_PepResultDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing results of PEP extraction.

The dataframe is expected to have *at least* the following columns:

* ``heartbeat_start_sample``: Start sample of segmented heartbeat
* ``heartbeat_end_sample``: End sample of segmented heartbeat
* ``r_peak_sample``: R-peak sample of segmented heartbeat
* ``q_peak_sample``: Q-peak sample of segmented heartbeat
* ``b_point_sample``: B-point sample of segmented heartbeat
* ``pep_sample``: Pre-ejection period (PEP) in samples
* ``pep_ms``: Pre-ejection period (PEP) in milliseconds



Additionally, the dataframe can contain the following columns:

* ``rr_interval_sample``: RR interval between the previous and the current heartbeat in samples
* ``rr_interval_ms``: RR interval between the previous and the current heartbeat in milliseconds
* ``heart_rate_bpm``: Heart rate in beats per minute, derived from RR interval
* ``nan_reason``: Reason why the PEP was set to NaN (e.g., "r_peak_nan", "no_zero_crossing")

"""


PhaseDict = dict[str, pd.DataFrame]
"""Dictionary containing general time-series data of **one single subject** split into **different phases**.

A ``PhaseDict`` is a dictionary with the following format:

{ "phase_1" : dataframe, "phase_2" : dataframe, ... }

Each ``dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* Index: :class:`pandas.DatetimeIndex` with timestamps, name of index level: ``time``

"""

HeartRatePhaseDict = dict[str, HeartRateDataFrame]
"""Dictionary containing time-series heart rate data of **one single subject** split into **different phases**.

A ``HeartRatePhaseDict`` is a dictionary with the following format:

{ "phase_1" : hr_dataframe, "phase_2" : hr_dataframe, ... }

Each ``hr_dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* ``time`` Index: :class:`pandas.DatetimeIndex` with heart rate sample timestamps
* ``Heart_Rate`` Column: heart rate values

"""

SubjectDataDict = dict[str, PhaseDict]
"""Dictionary representing time-series data from **multiple subjects** collected during a psychological protocol.

A ``SubjectDataDict`` is a nested dictionary with time-series data from multiple subjects, each containing data
from different phases. It is expected to have the level order `subject`, `phase`:

| {
|     "subject1" : { "phase_1" : dataframe, "phase_2" : dataframe, ... },
|     "subject2" : { "phase_1" : dataframe, "phase_2" : dataframe, ... },
|     ...
| }

This dictionary can, for instance, be rearranged to a :obj:`biopsykit.utils.dtypes.StudyDataDict`,
where the level order is reversed: `phase`, `subject`.
"""

HeartRateSubjectDataDict = Union[dict[str, HeartRatePhaseDict], dict[str, dict[str, HeartRatePhaseDict]]]
"""Dictionary with time-series heart rate data from **multiple subjects** collected during a psychological protocol.

A ``HeartRateSubjectDataDict`` is a nested dictionary with time-series heart rate data from multiple subjects,
each containing data from different phases. It is expected to have the level order `subject`, `phase`:

| {
|     "subject1" : { "phase_1" : hr_dataframe, "phase_2" : hr_dataframe, ... },
|     "subject2" : { "phase_1" : hr_dataframe, "phase_2" : hr_dataframe, ... },
|     ...
| }

Each ``hr_dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* ``time`` Index: :class:`pandas.DatetimeIndex` with heart rate sample timestamps
* ``Heart_Rate`` Column: heart rate values

This dictionary can, for instance, be rearranged to a :obj:`~biopsykit.utils.dtypes.HeartRateStudyDataDict`,
where the level order is reversed: `phase`, `subject`.

"""

StudyDataDict = dict[str, dict[str, pd.DataFrame]]
"""Dictionary with data from **multiple phases** collected during a psychological protocol.

A ``StudyDataDict`` is a nested dictionary with time-series data from multiple phases, each phase containing data
from different subjects. It is expected to have the level order `phase`, `subject`:

| {
|     "phase_1" : { "subject1" : dataframe, "subject2" : dataframe, ... },
|     "phase_2" : { "subject1" : dataframe, "subject2" : dataframe, ... },
|     ...
| }

This dict results from rearranging a :obj:`biopsykit.utils.dtypes.SubjectDataDict` by calling
:func:`~biopsykit.utils.data_processing.rearrange_subject_data_dict`.
"""


HeartRateStudyDataDict = dict[str, dict[str, HeartRateDataFrame]]
"""Dictionary with heart rate data from **multiple phases** collected during a psychological protocol.

A ``HeartRateStudyDataDict`` is a nested dictionary with time-series heart rate data from multiple phases,
each phase containing data from different subjects. It is expected to have the level order `phase`, `subject`:

| {
|     "phase_1" : { "subject1" : hr_dataframe, "subject2" : hr_dataframe, ... },
|     "phase_2" : { "subject1" : hr_dataframe, "subject2" : hr_dataframe, ... },
|     ...
| }

Each ``hr_dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* ``time`` Index: :class:`pandas.DatetimeIndex` with heart rate sample timestamps
* ``Heart_Rate`` Column: heart rate values

This dict results from rearranging a :obj:`~biopsykit.utils.dtypes.HeartRateSubjectDataDict` by calling
:func:`~biopsykit.utils.data_processing.rearrange_subject_data_dict`.
"""

MergedStudyDataDict = dict[str, pd.DataFrame]
"""Dictionary with merged time-series data of **multiple subjects**, split into **different phases**.

A ``MergedStudyDataDict`` is a dictionary with the following format:

| {
|     "phase_1" : merged_dataframe,
|     "phase_2" : merged_dataframe,
|     ...
| }

This dict results from merging the inner dictionary into one dataframe by calling
:func:`~biopsykit.utils.data_processing.merge_study_data_dict`.

.. note::
    Merging the inner dictionaries requires that the dataframes of all subjects have same length within each phase.

Each ``merged_dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* Index: time. Name of index level: ``time``
* Columns: time series data per subject, each subject has its own column.
  Name of the column index level: ``subject``
"""


def is_subject_condition_dataframe(
    data: SubjectConditionDataFrame, raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.SubjectConditionDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SubjectConditionDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SubjectConditionDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SubjectConditionDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SubjectConditionDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_multiindex(data, expected=False)
        _assert_has_index_levels(data, index_levels=["subject"], match_atleast=False, match_order=True)
        _assert_has_columns(data, [["condition"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SubjectConditionDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_subject_condition_dict(data: SubjectConditionDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.SubjectConditionDict`.

    Parameters
    ----------
    data : dict
        dict to check if it is a ``SubjectConditionDict``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SubjectConditionDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SubjectConditionDict``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SubjectConditionDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for val in data.values():
            _assert_is_dtype(val, (np.ndarray, list, pd.Index))
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SubjectConditionDict. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_codebook_dataframe(data: CodebookDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.CodebookDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``CodebookDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value


    Returns
    -------
    ``True`` if ``data`` is a ``CodebookDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)


    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``CodebookDataFrame``


    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.CodebookDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_index_levels(data, index_levels="variable", match_atleast=True, match_order=False)
        if not np.issubdtype(data.columns.dtype, np.integer):
            raise ValidationError(
                f"The dtypes of columns in a CodebookDataFrame are expected to be of type int, "
                f"but it is {data.columns.dtype}."
            )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a CodebookDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_mean_se_dataframe(data: MeanSeDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.MeanSeDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``MeanSeDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value


    Returns
    -------
    ``True`` if ``data`` is a ``MeanSeDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)


    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``MeanSeDataFrame``


    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.MeanSeDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        if data.columns.nlevels == 1:
            _assert_has_columns(data, [["mean", "se"]])
        else:
            _assert_has_columns_any_level(data, [["mean", "se"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a MeanSeDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_hr_phase_dict(data: HeartRatePhaseDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``HeartRatePhaseDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``HeartRatePhaseDict``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for df in data.values():
            is_heart_rate_dataframe(df)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a HeartRatePhaseDict. "
                f"The validation failed with the following error:\n\n{e!s}\n"
                "HeartRatePhaseDicts in an old format can be converted into the new format using "
                "`biopsykit.utils.legacy_helper.legacy_convert_hr_phase_dict()`"
            ) from e
        return False
    return True


def is_phase_dict(data: PhaseDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.dtypes.PhaseDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``PhaseDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``PhaseDict``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.PhaseDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for df in data.values():
            _assert_is_dtype(df, pd.DataFrame)
            _assert_has_multiindex(df, expected=False)
            _assert_has_column_multiindex(df, expected=False)
            _assert_has_index_levels(df, ["time"])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a PhaseDict. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_hr_subject_data_dict(data: HeartRateSubjectDataDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.dtypes.HeartRateSubjectDataDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``HeartRateSubjectDataDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``HeartRateSubjectDataDict``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.HeartRateSubjectDataDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for data_dict in data.values():
            is_hr_phase_dict(data_dict)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a HeartRateSubjectDataDict. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_study_data_dict(data: StudyDataDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.dtypes.StudyDataDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``StudyDataDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``StudyDataDict``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.StudyDataDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for data_dict in data.values():
            _assert_is_dtype(data_dict, dict)
            for df in data_dict.values():
                _assert_is_dtype(df, pd.DataFrame)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a StudyDataDict. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_subject_data_dict(data: SubjectDataDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.dtypes.SubjectDataDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SubjectDataDict```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SubjectDataDict``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SubjectDataDict``
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for data_dict in data.values():
            _assert_is_dtype(data_dict, dict)
            for df in data_dict.values():
                _assert_is_dtype(df, pd.DataFrame)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SubjectDataDict. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_merged_study_data_dict(data: MergedStudyDataDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.dtypes.MergedStudyDataDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``MergedStudyDataDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``MergedStudyDataDict``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.MergedStudyDataDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for df in data.values():
            _assert_is_dtype(df, pd.DataFrame)
            _assert_has_multiindex(df, expected=False)
            _assert_has_index_levels(df, ["time"])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a MergedStudyDataDict. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_biomarker_raw_dataframe(
    data: BiomarkerRawDataFrame, biomarker_type: Union[str, list[str]], raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SalivaRawDataFrame``
    saliva_type : str or list of str
        type of saliva data (or list of saliva types) in the dataframe, e.g., "cortisol" or "amylase"
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SalivaRawDataFrame```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SalivaRawDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        dataframe format

    """
    try:
        if biomarker_type is None:
            raise ValidationError("`saliva_type` is None!")
        if isinstance(biomarker_type, str):
            biomarker_type = [biomarker_type]
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_multiindex(data, nlevels=2, nlevels_atleast=True)
        _assert_has_index_levels(data, index_levels=["subject", "sample"], match_atleast=True, match_order=False)
        _assert_has_columns(data, [biomarker_type, [*biomarker_type, "time"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a BiomarkerRawDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_saliva_raw_dataframe(
    data: SalivaRawDataFrame, saliva_type: Union[str, list[str]], raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SalivaRawDataFrame``
    saliva_type : str or list of str
        type of saliva data (or list of saliva types) in the dataframe, e.g., "cortisol" or "amylase"
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SalivaRawDataFrame```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SalivaRawDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        dataframe format

    """
    return is_biomarker_raw_dataframe(data, saliva_type, raise_exception)


def is_saliva_feature_dataframe(
    data: SalivaFeatureDataFrame, saliva_type: str, raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SalivaFeatureDataFrame``
    saliva_type : str or list of str
        type of saliva data in the dataframe, e.g., "cortisol" or "amylase"
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SalivaFeatureDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SalivaFeatureDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame`
        dataframe format

    """
    try:
        if saliva_type is None:
            raise ValidationError("`saliva_type` is None!")
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_index_levels(data, index_levels="subject", match_atleast=True, match_order=False)
        # _assert_has_column_levels(data, column_levels="saliva_feature", match_atleast=True, match_order=False)
        _assert_has_column_prefix(data.columns, prefix=saliva_type)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SalivaFeatureDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_saliva_mean_se_dataframe(data: SalivaFeatureDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.SalivaMeanSeDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SalivaMeanSeDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SalivaMeanSeDataFrame```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SalivaMeanSeDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SalivaMeanSeDataFrame``
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_index_levels(data, index_levels="sample", match_atleast=True, match_order=False)
        _assert_has_columns(data, [["mean", "se"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SalivaMeanSeDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_sleep_endpoint_dataframe(data: SleepEndpointDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.SleepEndpointDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SleepEndpointDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SleepEndpointDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SleepEndpointDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SleepEndpointDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_is_dtype(data.index, pd.DatetimeIndex)
        _assert_has_index_levels(data, index_levels="date", match_atleast=True, match_order=False)
        _assert_has_columns(data, column_sets=[["sleep_onset", "wake_onset", "total_sleep_duration"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SleepEndpointDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_sleep_endpoint_dict(data: SleepEndpointDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dictionary is a :obj:`~biopsykit.utils.dtypes.SleepEndpointDict`.

    Parameters
    ----------
    data : dict
        data to check if it is a ``SleepEndpointDict``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SleepEndpointDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SleepEndpointDict``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SleepEndpointDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        expected_keys = ["date", "sleep_onset", "wake_onset", "total_sleep_duration"]
        if any(col not in list(data.keys()) for col in expected_keys):
            raise ValidationError(f"Not all of {expected_keys} are in the dictionary!")
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SleepEndpointDict. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_icg_raw_dataframe(data: IcgRawDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.IcgRawDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``IcgRawDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``IcgRawDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``IcgRawDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.IcgRawDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, column_sets=[["icg_der"], ["icg"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a IcgRawDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_ecg_raw_dataframe(data: EcgRawDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.EcgRawDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``EcgRawDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``EcgRawDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``EcgRawDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.EcgRawDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, column_sets=[["ecg"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a EcgRawDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_ecg_result_dataframe(data: EcgRawDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.EcgResultDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``EcgResultDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``EcgResultDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``EcgResultDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.EcgResultDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            column_sets=[
                ECG_RESULT_DATAFRAME_COLUMNS,
                ECG_RESULT_DATAFRAME_COLUMNS + HEART_RATE_DATAFRAME_COLUMNS,
            ],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a EcgResultDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_heart_rate_dataframe(data: HeartRateDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.HeartRateDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``HeartRateDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``HeartRateDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``HeartRateDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.HeartRateDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, column_sets=[HEART_RATE_DATAFRAME_COLUMNS])
        _assert_has_multiindex(data, expected=False)
        _assert_has_column_multiindex(data, expected=False)
        _assert_has_index_levels(data, ["time"])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a HeartRateDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_r_peak_dataframe(data: EcgRawDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.RPeakDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``RPeakDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``RPeakDataFrame```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``RPeakDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.RPeakDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            column_sets=[
                ["R_Peak_Idx", "RR_Interval"],
                ["R_Peak_Quality", "R_Peak_Idx", "RR_Interval"],
                R_PEAK_DATAFRAME_COLUMNS,
            ],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a RPeakDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_acc1d_dataframe(data: Acc3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.Acc1dDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``Acc1dDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``Acc1dDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``Acc1dDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.Acc1dDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            column_sets=[["acc"], ["acc_norm"]],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a Acc1dDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_acc3d_dataframe(data: Acc3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.Acc3dDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``Acc3dDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``Acc3dDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``Acc3dDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.Acc3dDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            column_sets=[
                ["acc_x", "acc_y", "acc_z"],
                [("acc", "x"), ("acc", "y"), ("acc", "z")],
            ],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a Acc3dDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_gyr1d_dataframe(data: Gyr3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.Gyr1dDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``Gyr1dDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``Gyr1dDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``Gyr1dDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.Gyr1dDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            column_sets=[["gyr"], ["gyr_norm"]],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a Gyr1dDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_gyr3d_dataframe(data: Gyr3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.Gyr3dDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``Gyr3dDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``Gyr3dDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``Gyr3dDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.Gyr3dDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            column_sets=[
                ["gyr_x", "gyr_y", "gyr_z"],
                [("gyr", "x"), ("gyr", "y"), ("gyr", "z")],
            ],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a Gyr3dDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_imu_dataframe(data: Gyr3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.ImuDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``ImuDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``ImuDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``ImuDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.ImuDataFrame`
        dataframe format

    """
    try:
        is_acc3d_dataframe(data, raise_exception=True)
        is_gyr3d_dataframe(data, raise_exception=True)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a ImuDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_sleep_wake_dataframe(data: SleepWakeDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.SleepWakeDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SleepWakeDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SleepWakeDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SleepWakeDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.SleepWakeDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, [["sleep_wake"]])
        if not all(data["sleep_wake"].between(0, 1, inclusive=True)):
            raise ValidationError(
                "Invalid values for sleep/wake prediction! Sleep/wake scores are expected to be "
                "in the interval [0, 1]."
            )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SleepWakeDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_heartbeat_segmentation_dataframe(
    data: HeartbeatSegmentationDataFrame, raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.HeartbeatSegmentationDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``HeartbeatSegmentationDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``HeartbeatSegmentationDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``HeartbeatSegmentationDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.HeartbeatSegmentationDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, [["start_sample", "end_sample", "r_peak_sample"]])
        # assert that columns with "_sample" in the end are of type int
        _assert_sample_columns_int(data)
        _assert_has_index_levels(data, "heartbeat_id", match_atleast=True, match_order=False)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a HeartbeatSegmentationDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_c_point_dataframe(data: CPointDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.CPointDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``CPointDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``CPointDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``CPointDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.CPointDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, [["c_point_sample"], ["c_point_sample", "nan_reason"]])
        # assert that columns with "_sample" in the end are of type int
        _assert_sample_columns_int(data)
        _assert_has_index_levels(data, "heartbeat_id", match_atleast=True, match_order=False)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a CPointDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_b_point_dataframe(data: BPointDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.BPointDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``BPointDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``BPointDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``BPointDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.BPointDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, [["b_point_sample"], ["b_point_sample", "nan_reason"]])
        # assert that columns with "_sample" in the end are of type int
        _assert_sample_columns_int(data)
        _assert_has_index_levels(data, "heartbeat_id", match_atleast=True, match_order=False)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a BPointDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_q_peak_dataframe(data: QPeakDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.QPeakDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``QPeakDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``QPeakDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``QPeakDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.QPeakDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, [["q_peak_sample"], ["q_peak_sample", "nan_reason"]])
        # assert that columns with "_sample" in the end are of type int
        _assert_sample_columns_int(data)
        _assert_has_index_levels(data, "heartbeat_id", match_atleast=True, match_order=False)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a QPeakDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True


def is_pep_result_dataframe(data: PepResultDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.dtypes.PepResultDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``PepResultDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``PepResultDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``PepResultDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.dtypes.PepResultDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            column_sets=[
                PEP_RESULT_DATAFRAME_COLUMNS,
                [*PEP_RESULT_DATAFRAME_COLUMNS, "rr_interval_sample", "rr_interval_ms", "heart_rate_bpm", "nan_reason"],
            ],
        )
        _assert_sample_columns_int(data)
        _assert_has_index_levels(data, "heartbeat_id", match_atleast=True, match_order=False)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a PepResultDataFrame. "
                f"The validation failed with the following error:\n\n{e!s}"
            ) from e
        return False
    return True



from typing import Optional

import numpy as np
import pandas as pd
from scipy.signal import butter, sosfiltfilt
from scipy.stats import median_abs_deviation
from tpcp import Algorithm

__all__ = ["BaseOutlierCorrection"]

from biopsykit.utils.dtypes import BPointDataFrame, CPointDataFrame


class BaseOutlierCorrection(Algorithm):
    """Base class for outlier correction algorithms for B-Point data."""

    _action_methods = "correct_outlier"

    points_: BPointDataFrame

    def correct_outlier(
        self,
        *,
        b_points: BPointDataFrame,
        c_points: Optional[CPointDataFrame],
        sampling_rate_hz: float,
        **kwargs,
    ):
        raise NotImplementedError("Method 'correct_outlier' must be implemented in a subclass!")

    @staticmethod
    def detect_b_point_outlier(stationary_data: pd.DataFrame) -> pd.DataFrame:
        """Detect outliers in stationary B-Point data.

        Outliers are detected based on the median absolute deviation of the stationary data. If the difference between
        the stationary data and the median is greater than 3 times the median absolute deviation, the data point is
        considered an outlier.

        Parameters
        ----------
        stationary_data : :class:`~pandas.DataFrame`
            DataFrame containing the stationary B-Point data

        Returns
        -------
        :class:`~pandas.DataFrame`
            DataFrame containing the detected outliers

        """
        median = np.nanmedian(stationary_data["statio_data"])
        median_abs_dev = median_abs_deviation(stationary_data["statio_data"], axis=0, nan_policy="propagate")
        outliers = pd.DataFrame(index=stationary_data.index, columns=["outliers"])
        outliers["outliers"] = False
        outliers.loc[(stationary_data["statio_data"] - median) > (3 * median_abs_dev), "outliers"] = True
        outliers.loc[stationary_data["statio_data"].isna(), "outliers"] = True

        return outliers[outliers["outliers"]]

    @staticmethod
    def stationarize_b_points(b_points: pd.DataFrame, c_points: pd.DataFrame, sampling_rate_hz: float) -> pd.DataFrame:
        """Stationarize B-Point data by removing the baseline.

        The B-points are stationarized by removing the baseline from the distance to the C-point. The baseline is
        estimated using a 4th order low-pass Butterworth filter with a cutoff frequency of 0.1 Hz.

        Parameters
        ----------
        b_points : :class:`~pandas.DataFrame`
            Dataframe containing the extracted B-Points per heartbeat, index functions as id of heartbeat
        c_points : :class:`~pandas.DataFrame`
            Dataframe containing the extracted C-Points per heartbeat, index functions as id of heartbeat
        sampling_rate_hz : float
            Sampling rate of ICG signal in Hz

        Returns
        -------
        :class:`~pandas.DataFrame`
            DataFrame containing the stationarized B-Point data

        """
        b_point_sample = b_points["b_point_sample"].astype(float)
        c_point_sample = c_points["c_point_sample"].astype(float)
        dist_to_c_point = ((c_point_sample - b_point_sample) / sampling_rate_hz).to_frame()
        dist_to_c_point.columns = ["dist_to_c_point_ms"]
        dist_to_c_point["b_point_sample"] = b_point_sample
        dist_to_c_point = dist_to_c_point.interpolate().ffill().bfill()

        sos = butter(4, Wn=0.1, btype="lowpass", output="sos", fs=1)

        # padlen for sosfiltfilt according to scipy documentation
        padlen = 3 * (2 * len(sos) + 1 - min((sos[:, 2] == 0).sum(), (sos[:, 5] == 0).sum()))
        if len(dist_to_c_point) <= padlen:
            # signal needs to be at least as long as padlen => reduce padlen for shorter signals
            padlen = len(dist_to_c_point) - 1
        baseline = sosfiltfilt(sos, dist_to_c_point["dist_to_c_point_ms"].values, padlen=padlen)

        # TODO Check if necessary after changing to sos? If yes, find out how it needs to be changed
        #  => changed by changing padlen

        statio_data = (dist_to_c_point["dist_to_c_point_ms"] - baseline).to_frame()
        statio_data.columns = ["statio_data"]
        statio_data["b_point_sample"] = dist_to_c_point["b_point_sample"]
        statio_data["baseline"] = baseline
        statio_data = statio_data.assign(padlen=padlen)

        return statio_data



"""Metrics for classification tasks."""

from inspect import getmembers

import numpy as np
import pandas as pd
import sklearn.metrics

from biopsykit.utils._types_internal import str_t


def _apply_score(row: pd.Series, score_func, pos_label: str):
    true_labels_folds = row[0]
    predicted_labels_folds = row[1]
    scores = [
        score_func(true_labels, predicted_labels, pos_label=pos_label)
        for true_labels, predicted_labels in zip(true_labels_folds, predicted_labels_folds)
    ]
    return pd.Series(scores)


def compute_additional_metrics(metric_summary: pd.DataFrame, metrics: str_t, pos_label: str):
    """Compute additional classification metrics from a ``SklearnPipelinePermuter`` metric summary dataframe.

    Parameters
    ----------
    metric_summary : :class:`~pandas.DataFrame`
        metric summary dataframe from :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter`
    metrics : list
        list of additional metrics to compute
    pos_label : str
        positive label

    """
    metric_slice = metric_summary[["true_labels_folds", "predicted_labels_folds"]].copy()
    metric_out = {}

    # ensure list
    if isinstance(metrics, str):
        metrics = [metrics]

    for metric in metrics:
        score_funcs = dict(getmembers(sklearn.metrics))
        if metric in score_funcs:
            score_func = score_funcs[f"{metric}"]
        elif f"{metric}_score" in score_funcs:
            score_func = score_funcs[f"{metric}_score"]
            metric = f"{metric}_score"  # noqa: PLW2901
        else:
            raise ValueError(f"Metric '{metric}' not found.")
        metric_out[metric] = metric_slice.apply(_apply_score, args=(score_func, pos_label), axis=1)
    metric_out = pd.concat(metric_out, names=["score", "folds"], axis=1)

    metric_out = metric_out.stack(["score", "folds"])
    metric_out = metric_out.groupby(metric_out.index.names[:-1]).agg(
        [("mean", lambda x: np.mean), ("std", lambda x: np.std(x))]  # noqa: ARG005
    )

    metric_out = metric_out.unstack("score").sort_index(axis=1, level="score")
    metric_out.columns = metric_out.columns.map("_test_".join)
    metric_summary = metric_summary.join(metric_out)

    return metric_summary



"""Extract features from static moments of IMU data."""

from typing import Optional, Union

import numpy as np
import pandas as pd
from scipy.stats import skew

from biopsykit.utils.array_handling import sanitize_input_nd
from biopsykit.utils.time import tz


def compute_features(
    data: pd.DataFrame,
    static_moments: pd.DataFrame,
    start: Optional[Union[str, pd.Timestamp]] = None,
    end: Optional[Union[str, pd.Timestamp]] = None,
    index: Optional[Union[int, str]] = None,
    timezone: Optional[str] = None,
) -> Optional[pd.DataFrame]:
    """Compute features based on frequency and duration of static moments in given input signal.

    This function computes the following features:

    * ``sm_number``: number of static moments in data
    * ``sm_max``: maximum duration of static moments, i.e., longest duration
    * ``sm_max_position``: location of the beginning of the longest static moment in the input data normalized to
      ``[0, 1]`` where 0 = ``start`` and 1 = ``end``
    * ``sm_median``: median duration of static moments
    * ``sm_mean``: mean duration of static moments
    * ``sm_std``: standard deviation of static moment durations
    * ``sm_skewness``: skewness of static moment durations

    The features are both computed on all detected static moments and on static moments that are longer than
    60 seconds (suffix ``_60``).


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        input data
    static_moments : :class:`~pandas.DataFrame`
        dataframe with beginning and end of static moments
    start : :class:`~pandas.Timestamp` or str, optional
        start timestamp in input data for feature extraction or ``None`` to set start index to the first index in
        ``data``. All samples *before* ``start`` will not be used for feature extraction.
    end : :class:`~pandas.Timestamp` or str, optional
        end timestamp in input data for feature extraction or ``None`` to set end index to the last index in
        ``data``. All samples *after* ``end`` will not be used for feature extraction.
    index : int or str, optional
        index label of the resulting dataframe or ``None`` to assign a default label (0)
    timezone : str, optional
        timezone of the recorded data or ``None`` to use default timezone ("Europe/Berlin")

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with extracted static moment features

    """
    if data.empty:
        return None

    start, end = _get_start_end(data, start, end, timezone)
    total_time = end - start

    static_moments = sanitize_input_nd(static_moments, ncols=2)

    durations = np.array([static_moment_duration(data, sequence) for sequence in static_moments])
    durations_60 = durations[durations >= 60]

    loc_max_moment = data.index[static_moments[np.argmax(durations)][0]]
    loc_max_moment_relative = (loc_max_moment - start) / total_time

    feature_dict = {"sm_max_position": loc_max_moment_relative}
    # feature_dict['sleep_bouts_number'.format(index)] = len(sleep_bouts)
    # feature_dict['wake_bouts_number'] = len(wake_bouts)

    # mean_orientations = mean_orientation(data, static_sequences)
    # dominant_orientation = mean_orientations.iloc[mean_orientations.index.argmax()]
    # dict_ori = {'sm_dominant_orientation_{}'.format(x): dominant_orientation.loc['acc_{}'.format(x)] for x
    #             in
    #             ['x', 'y', 'z']}
    # feature_dict.update(dict_ori)

    for dur, suffix in zip([durations, durations_60], ["", "_60"]):
        feature_dict[f"sm_number{suffix}"] = len(dur)
        feature_dict[f"sm_max{suffix}"] = np.max(dur)
        feature_dict[f"sm_median{suffix}"] = np.median(dur)
        feature_dict[f"sm_mean{suffix}"] = np.mean(dur)
        feature_dict[f"sm_std{suffix}"] = np.std(dur, ddof=1)
        feature_dict[f"sm_skewness{suffix}"] = skew(dur)

    if index is None:
        index = 0
    return pd.DataFrame(feature_dict, index=[index])


def _get_start_end(
    data: pd.DataFrame,
    start: Union[str, pd.Timestamp],
    end: Union[str, pd.Timestamp],
    timezone: str,
) -> tuple[Union[str, pd.Timestamp], Union[str, pd.Timestamp]]:
    if timezone is None:
        timezone = tz

    if start is None:
        start = data.index[0]
    if end is None:
        end = data.index[-1]

    start = _to_timestamp(start, timezone)
    end = _to_timestamp(end, timezone)
    return start, end


def _to_timestamp(date: Union[str, pd.Timestamp], timezone: str) -> pd.Timestamp:
    if isinstance(date, str):
        date = pd.Timestamp(date, tz=timezone)
    return date


def static_moment_duration(data: pd.DataFrame, start_end: np.array) -> float:
    """Compute duration of static moment.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        input data
    start_end : array
        start and end index of static moment to compute duration

    Returns
    -------
    float
        duration in seconds

    """
    return (data.index[start_end[1]] - data.index[start_end[0]]).total_seconds()


def mean_orientation(data: pd.DataFrame, static_moments: pd.DataFrame) -> pd.DataFrame:
    """Compute mean orientation of acceleration signal within static moment windows.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        input data
    static_moments : :class:`~pandas.DataFrame`
        dataframe with start and end indices of static moments

    Returns
    -------
    :class:`~pandas.DataFrame`
        mean orientation (x, y, z) of acceleration signal for each static moment window

    """
    static_moments = sanitize_input_nd(static_moments, 2)
    mean_orientations = [data.iloc[start_end[0] : start_end[1]] for start_end in static_moments]
    mean_orientations = {len(data): data.mean() for data in mean_orientations}
    mean_orientations = pd.DataFrame(mean_orientations).T
    # mean_orientations.rename(columns={'index': 'length'}, inplace=True)
    return mean_orientations



"""Module for generating Activity Counts from raw acceleration signals."""

import datetime
from typing import Optional, Union

import numpy as np
import pandas as pd
import pytz
from scipy import signal

from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.array_handling import add_datetime_index, downsample, sanitize_input_nd
from biopsykit.utils.dtypes import is_acc1d_dataframe, is_acc3d_dataframe
from biopsykit.utils.time import tz


class ActivityCounts:
    """Generate Activity Counts from raw acceleration signals.

    Actigraph Activity Counts are a unit used in many human activity studies.
    However, it can only be outputted by the official Actigraph Software.
    The following implementation uses a reverse engineered version of the Actigraph filter based on
    (Brønd et al., 2017).


    References
    ----------
    Brønd, J. C., Andersen, L. B., & Arvidsson, D. (2017). Generating ActiGraph Counts from Raw Acceleration Recorded
    by an Alternative Monitor. *Medicine and Science in Sports and Exercise*, 49(11), 2351-2360.
    https://doi.org/10.1249/MSS.0000000000001344

    """

    data: pd.DataFrame = None
    sampling_rate: float = None
    activity_counts_: np.ndarray = None
    timezone: datetime.tzinfo = tz

    def __init__(self, sampling_rate: float, timezone: Optional[str] = None):
        """Initialize a new ``ActivityCounts`` instance.

        Parameters
        ----------
        sampling_rate : float
            sampling rate of recorded data in Hz
        timezone: str
            timezone to which wear times will be converted

        """
        self.sampling_rate = sampling_rate
        if timezone:
            self.timezone = pytz.timezone(timezone)

    @staticmethod
    def _compute_norm(data: np.ndarray) -> np.ndarray:
        return np.linalg.norm(data, axis=1)

    @staticmethod
    def _aliasing_filter(data: np.ndarray, sampling_rate: Union[int, float]) -> np.ndarray:
        sos = signal.butter(5, [0.01, 7], "bp", fs=sampling_rate, output="sos")
        return signal.sosfiltfilt(sos, data)

    @staticmethod
    def _actigraph_filter(data: np.ndarray) -> np.ndarray:
        b = [
            0.04910898,
            -0.12284184,
            0.14355788,
            -0.11269399,
            0.05380374,
            -0.02023027,
            0.00637785,
            0.01851254,
            -0.03815411,
            0.04872652,
            -0.05257721,
            0.04784714,
            -0.04601483,
            0.03628334,
            -0.01297681,
            -0.00462621,
            0.01283540,
            -0.00937622,
            0.00344850,
            -0.00080972,
            -0.00019623,
        ]
        a = [
            1.00000000,
            -4.16372603,
            7.57115309,
            -7.98046903,
            5.38501191,
            -2.46356271,
            0.89238142,
            0.06360999,
            -1.34810513,
            2.47338133,
            -2.92571736,
            2.92983230,
            -2.78159063,
            2.47767354,
            -1.68473849,
            0.46482863,
            0.46565289,
            -0.67311897,
            0.41620323,
            -0.13832322,
            0.01985172,
        ]
        return signal.filtfilt(b, a, data)

    @staticmethod
    def _downsample(
        data: np.ndarray,
        sampling_rate: Union[int, float],
        final_sampling_rate: Union[int, float],
    ) -> np.ndarray:
        return downsample(data, sampling_rate, final_sampling_rate)

    @staticmethod
    def _truncate(data: np.ndarray) -> np.ndarray:
        upper_threshold = 2.13  # g
        lower_threshold = 0.068  # g
        data[data > upper_threshold] = upper_threshold
        data[data < lower_threshold] = 0
        return data

    @staticmethod
    def _digitize_8bit(data: np.ndarray) -> np.ndarray:
        max_val = 2.13  # g
        data //= max_val / (2**7)
        return data

    @staticmethod
    def _accumulate_second_bins(data: np.ndarray) -> np.ndarray:
        n_samples = 10
        #  Pad data at end to "fill" last bin
        padded_data = np.pad(data, (0, n_samples - len(data) % n_samples), "constant", constant_values=0)
        return padded_data.reshape((len(padded_data) // n_samples, -1)).sum(axis=1)

    def calculate(self, data: arr_t) -> arr_t:
        """Calculate Activity Counts from acceleration data.

        Parameters
        ----------
        data : array_like
            input data. Must either be 3-d or 1-d (e.g., norm, or a specific axis) acceleration data

        Returns
        -------
        array_like
            output data with Activity Counts

        """
        start_idx = None
        if isinstance(data, pd.DataFrame):
            # if dataframe, assert to be a acceleration dataframe according to biopsykit's convention
            if data.shape[1] == 3:
                is_acc3d_dataframe(data)
            if data.shape[1] == 1:
                is_acc1d_dataframe(data)
            data = data.filter(like="acc")
            if isinstance(data.index, pd.DatetimeIndex):
                start_idx = data.index[0]

        arr = sanitize_input_nd(data, ncols=(1, 3))

        if arr.shape[1] not in (1, 3):
            raise ValueError(
                f"{self.__class__.__name__} takes only 1-d or 3-d accelerometer data! Got {arr.shape[1]}-d data."
            )
        if arr.shape[1] != 1:
            arr = self._compute_norm(arr)

        arr = self._downsample(arr, self.sampling_rate, 30)
        arr = self._aliasing_filter(arr, 30)
        arr = self._actigraph_filter(arr)
        arr = self._downsample(arr, 30, 10)
        arr = np.abs(arr)
        arr = self._truncate(arr)
        arr = self._digitize_8bit(arr)
        arr = self._accumulate_second_bins(arr)
        if start_idx is not None:
            arr = add_datetime_index(arr, start_idx, 1, column_name=["activity_counts"])

        return arr



import pandas as pd

__all__ = ["BaseEcgExtraction"]

from biopsykit.signals._base_extraction import BaseExtraction


class BaseEcgExtraction(BaseExtraction):
    """Base class for ECG event extraction algorithms."""

    def extract(
        self,
        *,
        ecg: pd.DataFrame,
        heartbeats: pd.DataFrame,
        sampling_rate_hz: float,
    ):
        """Extract events from ECG signal.

        This is an abstract method that needs to be implemented in a subclass.

        Parameters
        ----------
        ecg : :class:`~pandas.DataFrame`
            ECG signal
        heartbeats : :class:`~pandas.DataFrame`
            DataFrame containing segmented heartbeats. Each row contains start, end, and R-peak location (in samples
            from beginning of signal) of that heartbeat, index functions as id of heartbeat
        sampling_rate_hz : float
            Sampling rate of ECG signal in Hz

        Returns
        -------
        self

        Raises
        ------
        NotImplementedError
            If this method is called from the base class

        """
        raise NotImplementedError("This is an abstract method and needs to be implemented in a subclass.")



"""Module with classes representing different psychological protocols."""

from biopsykit.protocols import plotting
from biopsykit.protocols.base import BaseProtocol
from biopsykit.protocols.car import CAR
from biopsykit.protocols.cft import CFT
from biopsykit.protocols.mist import MIST
from biopsykit.protocols.tsst import TSST

__all__ = ["CAR", "CFT", "MIST", "TSST", "BaseProtocol", "plotting"]



import neurokit2 as nk
import pandas as pd
from tpcp import Algorithm, Parameter

__all__ = ["EcgPreprocessingNeurokit"]

from biopsykit.utils.array_handling import sanitize_input_dataframe_1d
from biopsykit.utils.dtypes import EcgRawDataFrame, is_ecg_raw_dataframe


class EcgPreprocessingNeurokit(Algorithm):
    """ECG preprocessing algorithm using NeuroKit2 [1]_.

    This class provides an interface to the ECG preprocessing functions of NeuroKit2.

    References
    ----------
    .. [1] Makowski, D., Pham, T., Lau, Z. J., Brammer, J. C., Lesspinasse, F., Pham, H., Schölzel, C., & S.H. Chen
        (2021). NeuroKit2: A Python Toolbox for Neurophysiological Signal Processing. Behavior Research Methods.
        https://doi.org/10.3758/s13428-020-01516-y

    """

    _action_methods = "clean"

    method: Parameter[str]

    ecg_clean_: pd.DataFrame

    def __init__(self, method: str = "biosppy"):
        """Initialize ``EcgPreprocessingNeurokit`` instance.

        Parameters
        ----------
        method : str, optional
            Cleaning method to use. Options are:
                * "biosppy" (the default): Use the preprocessing parameters provided by
                `biosppy <https://biosppy.readthedocs.io/en/stable/>`_ library for cleaning. It uses an FIR filter
                with cut-off frequencies of [0.67, 45] Hz and order = 1.5 * sampling_rate.
                * "neurokit": Use the `NeuroKit2 <https://neurokit2.readthedocs.io/en/latest/>`_ library for cleaning.

        """
        self.method = method

    def clean(self, *, ecg: EcgRawDataFrame, sampling_rate_hz: int):
        """Clean ECG signal using :func:`~neurokit2.ecg_clean`.

        Parameters
        ----------
        ecg : :class:`~pandas.DataFrame`
            ECG signal
        sampling_rate_hz : int
            Sampling rate of the ECG signal in Hz

        Returns
        -------
        self

        Raises
        ------
        ValueError
            If the specified method is not implemented yet

        """
        is_ecg_raw_dataframe(ecg)
        ecg = sanitize_input_dataframe_1d(ecg, column="ecg")
        ecg = ecg.squeeze()
        if ecg.empty:
            self.ecg_clean_ = pd.DataFrame(index=ecg.index, columns=["ecg"])
            return self

        if self.method not in ["neurokit", "biosppy"]:
            raise ValueError("Not implemented yet!")
        clean_signal = nk.ecg_clean(ecg, sampling_rate=sampling_rate_hz, method=self.method)
        self.ecg_clean_ = pd.DataFrame(clean_signal, index=ecg.index, columns=["ecg"])

        return self



"""Internal helpers for dataset validation."""

from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np
import pandas as pd

from biopsykit.utils._types_internal import _Hashable, path_t
from biopsykit.utils.exceptions import FileExtensionError, ValidationError, ValueRangeError


def _assert_is_dir(path: path_t, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check if a path is a directory.

    Parameters
    ----------
    path : path or str
        path to check if it's a directory
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``path`` is a directory, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValueError
        if ``raise_exception`` is ``True`` and ``path`` is not a directory

    """
    # ensure pathlib
    file_name = Path(path)
    if not file_name.is_dir():
        if raise_exception:
            raise ValueError(f"The path '{path}' is expected to be a directory, but it's not!")
        return False

    return True


def _assert_file_extension(
    file_name: path_t, expected_extension: Union[str, Sequence[str]], raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check if a file has the correct file extension.

    Parameters
    ----------
    file_name : path or str
        file name to check for correct extension
    expected_extension : str or list of str
        file extension (or a list of file extensions) to check for
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``file_name`` ends with one of the specified file extensions, ``False`` otherwise
    (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.FileExtensionError`
        if ``raise_exception`` is ``True`` and ``file_name`` does not end with any of the specified
        ``expected_extension``

    """
    # ensure pathlib
    file_name = Path(file_name)
    if isinstance(expected_extension, str):
        expected_extension = [expected_extension]
    if file_name.suffix not in expected_extension:
        if raise_exception:
            raise FileExtensionError(
                f"The file name extension is expected to be one of {expected_extension}. "
                f"Instead it has the following extension: {file_name.suffix}"
            )
        return False
    return True


def _assert_is_dtype(
    obj, dtype: Union[type, tuple[type, ...]], raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check if an object has a specific data type.

    Parameters
    ----------
    obj : any object
        object to check
    dtype : type or list of type
        data type of tuple of data types to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``obj`` is one of the expected data types, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``obj`` is none of the expected data types

    """
    if not isinstance(obj, dtype):
        if raise_exception:
            raise ValidationError(f"The data object is expected to be one of ({dtype},). But it is a {type(obj)}")
        return False
    return True


def _assert_has_multiindex(
    df: pd.DataFrame,
    expected: Optional[bool] = True,
    nlevels: Optional[int] = 2,
    nlevels_atleast: Optional[int] = False,
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    """Check if a :any:`pandas.DataFrame` has a :any:`pandas.MultiIndex` as index.

    Parameters
    ----------
    df : :class:`~pandas.DataFrame`
        The dataframe to check
    expected : bool, optional
        Whether the df is expected to have a MultiIndex index or not
    nlevels : int, optional
        If MultiIndex is expected, how many levels the MultiIndex index should have
    nlevels_atleast : bool, optional
        Whether the MultiIndex has to have at least ``nlevels`` (``True``)
        or exactly match the number of levels (``False``)
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``df`` meets the expected index format, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``df`` does not meet the expected index format

    """
    return _multiindex_check_helper(
        df=df,
        idx_or_col="index",
        expected=expected,
        nlevels=nlevels,
        nlevels_atleast=nlevels_atleast,
        raise_exception=raise_exception,
    )


def _assert_has_index_levels(
    df: pd.DataFrame,
    index_levels: Iterable[_Hashable],
    match_atleast: Optional[bool] = False,
    match_order: Optional[bool] = False,
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    """Check if the dataframe has all index level names.

    Parameters
    ----------
    df : :class:`~pandas.DataFrame`
        The dataframe to check
    index_levels : list
        Set of index level names to check
    match_atleast : bool, optional
        Whether the MultiIndex columns have to have at least the specified column levels (``True``)
        or exactly match the column levels (``False``)
    match_order : bool, optional
        Whether to also match the level order
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``df`` has the expected index level names, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``df`` does not have the expected index level names

    """
    return _multiindex_level_names_helper(
        df,
        level_names=index_levels,
        idx_or_col="index",
        match_atleast=match_atleast,
        match_order=match_order,
        raise_exception=raise_exception,
    )


def _assert_has_columns(
    df: pd.DataFrame,
    column_sets: Sequence[Union[list[_Hashable], list[str], pd.Index]],
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    """Check if the dataframe has at least all columns sets.

    Parameters
    ----------
    df : :class:`~pandas.DataFrame`
        The dataframe to check
    column_sets : list
        Column set or list of column sets to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``df`` has the expected column names, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``df`` does not have the expected index level names

    Examples
    --------
    >>> df = pd.DataFrame()
    >>> df.columns = ["col1", "col2"]
    >>> _assert_has_columns(df, [["other_col1", "other_col2"], ["col1", "col2"]])
    >>> # This raises no error, as df contains all columns of the second set

    """
    columns = df.columns
    result = False
    for col_set in column_sets:
        result = result or all(v in columns for v in col_set)

    if result is False:
        if len(column_sets) == 1:
            helper_str = f"the following columns: {column_sets[0]}"
        else:
            helper_str = f"one of the following sets of columns: {column_sets}"
        if raise_exception:
            raise ValidationError(
                f"The dataframe is expected to have {helper_str}. "
                f"Instead it has the following columns: {list(df.columns)}"
            )
    return result


def _assert_has_column_multiindex(
    df: pd.DataFrame,
    expected: Optional[bool] = True,
    nlevels: Optional[int] = 2,
    nlevels_atleast: Optional[int] = False,
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    """Check if a :any:`pandas.DataFrame` has a :any:`pandas.MultiIndex` as columns.

    Parameters
    ----------
    df : :class:`~pandas.DataFrame`
        The dataframe to check
    expected : bool, optional
        Whether the df is expected to have MultiIndex column or not
    nlevels : int, optional
        If MultiIndex is expected, how many levels the MultiIndex columns should have
    nlevels_atleast : bool, optional
        Whether the MultiIndex has to have at least ``nlevels`` (``True``)
        or exactly match the number of levels (``False``)
    raise_exception : bool, optional
        Whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``df`` meets the expected column index format, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception` is ``True`` and ``df`` does not meet the expected column index format

    """
    return _multiindex_check_helper(
        df=df,
        idx_or_col="column",
        expected=expected,
        nlevels=nlevels,
        nlevels_atleast=nlevels_atleast,
        raise_exception=raise_exception,
    )


def _assert_has_columns_any_level(
    df: pd.DataFrame,
    columns_sets: Sequence[Union[list[_Hashable], list[str], pd.Index]],
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    """Check if the dataframe has the expected set of column names at any level of a :any:`pandas.MultiIndex`.

    Parameters
    ----------
    df : :class:`~pandas.DataFrame`
        The dataframe to check
    columns_sets : list
        Column set of list of column sets to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``df`` has the expected column names at any :any:`pandas.MultiIndex` level,
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``df`` does not have the expected column names

    Examples
    --------
    >>> df = pd.DataFrame()
    >>> df.columns = pd.MultiIndex.from_tuples([("Name", "col1"), ("Name", "col2")])
    >>> _assert_has_columns_any_level(df, [["col1", "col2"]])
    >>> # This raises no error, as df contains all columns in the seconds level

    """
    _assert_has_column_multiindex(df, expected=True, nlevels_atleast=True)
    column_levels = [np.array(df.columns.get_level_values(i)) for i in range(df.columns.nlevels)]
    result = False
    for columns in column_levels:
        for col_set in columns_sets:
            result = result or all(v in columns for v in col_set)

    if result is False:
        if len(columns_sets) == 1:
            helper_str = f"the following columns: {columns_sets[0]}"
        else:
            helper_str = f"one of the following sets of columns: {columns_sets}"
        if raise_exception:
            raise ValidationError(
                f"The dataframe is expected to have {helper_str} at any level of the MultiIndex. Instead it has the "
                f"following MultiIndex columns: {column_levels}"
            )
    return result


def _assert_has_column_levels(
    df: pd.DataFrame,
    column_levels: Iterable[_Hashable],
    match_atleast: Optional[bool] = False,
    match_order: Optional[bool] = False,
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    """Check if the dataframe has all column level names of a MultiIndex column.

    Parameters
    ----------
    df : :class:`~pandas.DataFrame`
        The dataframe to check
    column_levels : list
        Set of column level names to check
    match_atleast : bool, optional
        Whether the MultiIndex columns have to have at least the specified column levels (``True``)
        or exactly match the column levels (``False``)
    match_order : bool, optional
        Whether to also match the level order
    raise_exception : bool, optional
        Whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``df`` has the expected column level names, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``df`` does not have the expected index level names

    """
    return _multiindex_level_names_helper(
        df,
        level_names=column_levels,
        idx_or_col="column",
        match_atleast=match_atleast,
        match_order=match_order,
        raise_exception=raise_exception,
    )


def _assert_value_range(
    data: Union[pd.DataFrame, pd.Series],
    value_range: Sequence[Union[int, float]],
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    """Check if all values are within the specified range.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check values
    value_range : tuple of numbers
        value range in the format [min_val, max_val]
    raise_exception : bool, optional
        Whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if all values in ``data`` are within ``value_range``, ``False`` otherwise
    (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValueRangeError`
        if ``raise_exception`` is ``True`` and any value of ``data`` is not within ``value_range``

    """
    max_val = np.nanmax(data)
    min_val = np.nanmin(data)
    if not (min_val >= value_range[0] and max_val <= value_range[1]):
        if raise_exception:
            raise ValueRangeError(
                "Some of the values are out of the expected range. "
                f"Expected were values in the range {value_range}, got values in the range {[min_val, max_val]}. "
                "If values are part of questionnaire scores, "
                "you can convert questionnaire items into the correct range by calling "
                "`biopsykit.questionnaire.utils.convert_scale()`."
            )
        return False
    return True


def _assert_num_columns(
    data: pd.DataFrame, num_cols: Union[int, Sequence[int]], raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check if dataframe has (any of) the required number of columns.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check
    num_cols : int or list of int
        the required number of columns (or any of the required number of columns in case ``num_cols`` is a list)
    raise_exception : bool, optional
        Whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` has the required number of columns, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``data`` does not have the required number of columns

    """
    if isinstance(num_cols, int):
        num_cols = [num_cols]

    if not any(len(data.columns) == num for num in num_cols):
        if raise_exception:
            raise ValidationError(
                "The dataframe does not have the required number of columns. "
                f"Expected were any of {num_cols} columns, but has {len(data.columns)} columns."
            )
        return False
    return True


def _assert_len_list(data: Sequence, length: int, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check if a list has the required length.

    Parameters
    ----------
    data : list
        list to check
    length : int
        the required length or the list
    raise_exception : bool, optional
        Whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` has the required length, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``data`` does not have the required length

    """
    _assert_is_dtype(data, (list, tuple, np.ndarray))
    if len(data) != length:
        if raise_exception:
            raise ValidationError(
                "The list does not have the required length. "
                f"Expected was length {length}, but it has length {len(data)}."
            )
        return False
    return True


def _assert_dataframes_same_length(
    df_list: Sequence[pd.DataFrame], raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check if all dataframes have same length.

    Parameters
    ----------
    df_list : list
        list of dataframes to check
    raise_exception : bool, optional
        Whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if all dataframes in ``df_list`` have same length, ``False`` otherwise
    (if ``raise_exception`` is ``False``)

    Raises
    ------
    :exc:`~biopsykit.exceptions.ValidationError`
        if ``raise_exception`` is ``True`` and ``data`` does not have the required length

    """
    if len({len(df) for df in df_list}) != 1:
        if raise_exception:
            raise ValidationError("Not all dataframes have the same length!")
        return False
    return True


def _assert_sample_columns_int(data: pd.DataFrame) -> None:
    """Assert that the columns of a DataFrame that have "_sample" in their name are of type int."""
    if not any(data.columns.str.contains("_sample")):
        raise ValidationError("DataFrame does not contain any columns with '_sample' in their name!")
    for col in data.columns:
        if "_sample" in col and not pd.api.types.is_integer_dtype(data[col]):
            raise ValidationError(f"Column '{col}' is not of type 'int'!")


def _multiindex_level_names_helper_get_expected_levels(
    ac_levels: Sequence[str],
    ex_levels: Sequence[str],
    match_atleast: Optional[bool] = False,
    match_order: Optional[bool] = False,
) -> bool:
    if match_order:
        if match_atleast:
            ac_levels_slice = ac_levels[: len(ex_levels)]
            expected = ex_levels == ac_levels_slice
        else:
            expected = ex_levels == ac_levels
    elif match_atleast:
        expected = all(level in ac_levels for level in ex_levels)
    else:
        expected = sorted(ex_levels) == sorted(ac_levels)

    return expected


def _multiindex_level_names_helper(
    df: pd.DataFrame,
    level_names: Iterable[_Hashable],
    idx_or_col: str,
    match_atleast: Optional[bool] = False,
    match_order: Optional[bool] = False,
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    if isinstance(level_names, str):
        level_names = [level_names]

    ex_levels = list(level_names)
    ac_levels = list(df.index.names) if idx_or_col == "index" else list(df.columns.names)

    expected = _multiindex_level_names_helper_get_expected_levels(ac_levels, ex_levels, match_atleast, match_order)

    if not expected:
        if raise_exception:
            raise ValidationError(
                f"The dataframe is expected to have exactly the following {idx_or_col} level names {level_names}, "
                f"but it has {ac_levels}"
            )
        return False
    return True


def _multiindex_check_helper(
    df: pd.DataFrame,
    idx_or_col: str,
    expected: Optional[bool] = True,
    nlevels: Optional[int] = 2,
    nlevels_atleast: Optional[int] = False,
    raise_exception: Optional[bool] = True,
) -> Optional[bool]:
    has_multiindex, nlevels_act = _multiindex_check_helper_get_levels(df, idx_or_col)

    if has_multiindex is not expected:
        return _multiindex_check_helper_not_expected(idx_or_col, nlevels, nlevels_act, expected, raise_exception)

    if has_multiindex is True:
        expected = nlevels_act >= nlevels if nlevels_atleast else nlevels_act == nlevels
        if not expected:
            if raise_exception:
                raise ValidationError(
                    f"The dataframe is expected to have a MultiIndex with {nlevels} {idx_or_col} levels. "
                    f"But it has a MultiIndex with {nlevels_act} {idx_or_col} levels."
                )
            return False
    return True


def _multiindex_check_helper_get_levels(df: pd.DataFrame, idx_or_col: str) -> tuple[bool, int]:
    if idx_or_col == "index":
        has_multiindex = isinstance(df.index, pd.MultiIndex)
        nlevels_act = df.index.nlevels
    else:
        has_multiindex = isinstance(df.columns, pd.MultiIndex)
        nlevels_act = df.columns.nlevels

    return has_multiindex, nlevels_act


def _multiindex_check_helper_not_expected(
    idx_or_col: str, nlevels: int, nlevels_act: int, expected: bool, raise_exception: bool
) -> Optional[bool]:
    if not expected:
        if raise_exception:
            raise ValidationError(
                f"The dataframe is expected to have a single level as {idx_or_col}. "
                f"But it has a MultiIndex with {nlevels_act} {idx_or_col} levels."
            )
        return False
    if raise_exception:
        raise ValidationError(
            f"The dataframe is expected to have a MultiIndex with {nlevels} {idx_or_col} levels. "
            f"It has just a single normal {idx_or_col} level."
        )
    return False


def _assert_has_column_prefix(
    columns: Sequence[str], prefix: str, raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether all columns start with the same prefix.

    Parameters
    ----------
    columns : list of str
        list of column names
    prefix : str
        expected prefix of all columns
    raise_exception : bool, optional
        Whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``columns`` all start with ``prefix``, ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and one of ``columns`` is not a string or does not start with ``prefix``

    """
    if prefix is None or len(prefix) == 0:
        if raise_exception:
            raise ValidationError("'prefix' is None or empty!")
        return False

    for col in columns:
        return _check_has_column_prefix_single_col(columns, col, prefix, raise_exception)

    return True


def _check_has_column_prefix_single_col(
    columns: Sequence[str], col: Any, prefix: str, raise_exception: bool
) -> Optional[bool]:
    if not _assert_is_dtype(col, str, raise_exception=False):
        if raise_exception:
            raise ValidationError(f"Column '{col}' from {columns} is not a string!")
        return False
    if not col.startswith(prefix):
        if raise_exception:
            raise ValidationError(f"Column '{col}' from {columns} are starting with the required prefix '{prefix}'!")
        return False
    return True



"""Utility functions for working with saliva dataframes."""

import re
from collections.abc import Sequence
from datetime import datetime, time
from typing import Optional, Union

import numpy as np
import pandas as pd

from biopsykit.utils._datatype_validation_helper import _assert_has_index_levels, _assert_is_dtype
from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.dtypes import SalivaFeatureDataFrame, SalivaRawDataFrame, _SalivaRawDataFrame

__all__ = [
    "extract_saliva_columns",
    "get_saliva_column_suggestions",
    "saliva_feature_wide_to_long",
    "sample_times_datetime_to_minute",
]


def saliva_feature_wide_to_long(
    data: SalivaFeatureDataFrame,
    saliva_type: str,
) -> pd.DataFrame:
    """Convert ``SalivaFeatureDataFrame`` from wide-format into long-format.

    Parameters
    ----------
    data : :class:`~biopsykit.utils.dtypes.SalivaFeatureDataFrame`
        dataframe containing saliva features in wide-format, i.e. one column per saliva sample, one row per subject.
    saliva_type : str
        saliva type (e.g. 'cortisol')

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with saliva features in long-format

    """
    data = data.filter(like=saliva_type)
    index_cols = list(data.index.names)
    j = "saliva_feature"
    # iteratively build up long-format dataframe
    data = pd.wide_to_long(
        data.reset_index(),
        stubnames=saliva_type,
        i=index_cols,
        j=j,
        sep="_",
        suffix=r"\w+",
    )

    # reorder levels and sort
    return data.reorder_levels([*index_cols, j]).sort_index()


def get_saliva_column_suggestions(data: pd.DataFrame, saliva_type: Union[str, Sequence[str]]) -> Sequence[str]:
    """Automatically extract possible saliva data columns from a pandas dataframe.

    This is for example useful when one large dataframe is used to store demographic information,
    questionnaire data and saliva data.

    Parameters
    ----------
    data: :class:`~pandas.DataFrame`
        dataframe which should be extracted
    saliva_type: str or list of str
        saliva type variable which or list of saliva types should be used to extract columns (e.g. 'cortisol')

    Returns
    -------
    list or dict
        list of suggested columns containing saliva data or dict of such if ``saliva_type`` is a list

    """
    # check if input is dataframe
    _assert_is_dtype(data, pd.DataFrame)

    if isinstance(saliva_type, list):
        dict_result = {}
        for saliva in saliva_type:
            dict_result[saliva] = get_saliva_column_suggestions(data=data, saliva_type=saliva)
        return dict_result

    if saliva_type not in _dict_saliva_type_suggs:
        raise ValueError(f"Invalid saliva type '{saliva_type}'! Must be one of {list(_dict_saliva_type_suggs.keys())}.")

    sugg_filt = list(
        filter(
            lambda col: any(k in col for k in _dict_saliva_type_suggs[saliva_type]),
            data.columns,
        )
    )
    sugg_filt = list(filter(lambda s: any(str(i) in s for i in range(20)), sugg_filt))
    sugg_filt = list(
        filter(
            lambda s: all(
                k not in s
                for k in (
                    "AUC",
                    "auc",
                    "TSST",
                    "max",
                    "log",
                    "inc",
                    "lg",
                    "ln",
                    "GenExp",
                    "inv",
                )
            ),
            sugg_filt,
        )
    )
    # replace il{} with il6 since this was removed out by the previous filter operation
    sugg_filt = [re.sub(r"\d", "{}", s).replace("il{}", "il6").replace("IL{}", "IL6") for s in sugg_filt]
    sugg_filt = sorted(filter(lambda s: "{}" in s, set(sugg_filt)))

    # build regex for column extraction
    sugg_filt = ["^{}$".format(s.replace("{}", r"(\d)")) for s in sugg_filt]
    return sugg_filt


def extract_saliva_columns(
    data: pd.DataFrame, saliva_type: Union[str, Sequence[str]], col_pattern: Optional[Union[str, Sequence[str]]] = None
) -> Union[pd.DataFrame, dict[str, pd.DataFrame]]:
    """Extract saliva sample columns from a pandas dataframe.

    Parameters
    ----------
    data: :class:`~pandas.DataFrame`
        dataframe to extract columns from
    saliva_type: str or list of str
        saliva type variable or list of saliva types which should be used to extract columns (e.g. 'cortisol')
    col_pattern: str, optional
        string pattern or list of string patterns to identify saliva columns.
        If ``None``, it is attempted to automatically infer column names using :func:`get_saliva_column_suggestions()`.
        If ``col_pattern`` is a list, it must be the same length like ``saliva_type``.

    Returns
    -------
    :class:`~pandas.DataFrame` or dict
        pandas dataframe with extracted columns or dict of such if ``saliva_type`` is a list

    """
    if isinstance(saliva_type, list):
        if isinstance(col_pattern, list) and len(saliva_type) is not len(col_pattern):
            raise ValueError("'saliva_type' and 'col_pattern' must have same length!")
        dict_result = {}
        if col_pattern is None:
            col_pattern = [None] * len(saliva_type)
        for saliva, col_p in zip(saliva_type, col_pattern):
            dict_result[saliva] = extract_saliva_columns(data=data, saliva_type=saliva, col_pattern=col_p)
        return dict_result

    if col_pattern is None:
        col_suggs = get_saliva_column_suggestions(data, saliva_type)
        if len(col_suggs) > 1:
            raise ValueError(
                "More than one possible column pattern was found! "
                f"Please check manually which pattern is correct: {col_suggs}"
            )
        col_pattern = col_suggs[0]
    return data.filter(regex=col_pattern)


def _sample_times_datetime_to_minute_apply(
    sample_times: Union[pd.DataFrame, pd.Series],
) -> Union[pd.DataFrame, pd.Series]:
    if isinstance(sample_times.to_numpy().flatten()[0], (pd.Timedelta, np.timedelta64)):
        return sample_times.apply(pd.to_timedelta)
    return sample_times.astype(str).apply(pd.to_datetime)


def sample_times_datetime_to_minute(sample_times: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """Convert sample times from datetime or timedelta objects into minutes.

    In order to compute certain saliva features (such as :func:`~biopsykit.saliva.auc` or
    :func:`~biopsykit.saliva.slope`) the saliva sampling times are needed.
    This function can be used to convert sampling times into minutes relative to the first saliva sample.

    Parameters
    ----------
    sample_times : :class:`~pandas.Series` or :class:`~pandas.DataFrame`
        saliva sampling times in a Python datetime- or timedelta-related format.
        If ``sample_times`` is a Series, it is assumed to be in long-format and will be unstacked into wide-format
        along the `sample` level.
        If ``sample_times`` is a DataFrame, it is assumed to be in wide-format already.
        If values in ``sample_times`` are ``str``, they are assumed to be strings with time information only
        (**not** including date), e.g., "09:00", "09:15", ...

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe in wide-format with saliva sampling times in minutes relative to the first saliva sample

    Raises
    ------
    ValueError
        if sample times are not in a datetime- or timedelta-related format

    """
    if isinstance(sample_times.to_numpy().flatten()[0], str):
        sample_times = _get_sample_times_str(sample_times)

    if not isinstance(
        sample_times.to_numpy().flatten()[0], (time, datetime, pd.Timedelta, np.timedelta64, np.datetime64)
    ):
        raise TypeError(
            "Sample times must be instance of `datetime.datetime()`, `datetime.time()`,"
            " `np.datetime64`, `np.timedelta64`, or `pd.Timedelta`!"
        )

    is_series = isinstance(sample_times, pd.Series)
    if is_series:
        _assert_has_index_levels(sample_times, index_levels=["sample"], match_atleast=True)
        # unstack the multi-index dataframe in the 'samples' level so that time differences can be computed in minutes.
        # Then stack it back together
        sample_times = sample_times.unstack(level="sample")

    sample_times = _sample_times_datetime_to_minute_apply(sample_times)

    sample_times = sample_times.diff(axis=1).apply(lambda s: (s.dt.total_seconds() / 60))
    sample_times = sample_times.cumsum(axis=1)
    sample_times.iloc[:, 0] = sample_times.iloc[:, 0].fillna(0)
    if is_series:
        sample_times = sample_times.stack()
    return sample_times


def _get_sample_times_str(sample_times: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    if isinstance(sample_times, pd.DataFrame):
        return pd.to_timedelta(sample_times.stack()).unstack("sample")
    return pd.to_timedelta(sample_times)


def _remove_s0(data: SalivaRawDataFrame) -> SalivaRawDataFrame:
    """Remove first saliva sample.

    Parameters
    ----------
    data : :class:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format

    Returns
    -------
    :class:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format without the first saliva sample

    """
    data = data.drop(0, level="sample", errors="ignore")
    data = data.drop("0", level="sample", errors="ignore")
    data = data.drop("S0", level="sample", errors="ignore")
    return _SalivaRawDataFrame(data)


def _check_sample_times(sample_times: np.array) -> None:
    """Check that all sample times are monotonously increasing.

    Parameters
    ----------
    sample_times : array-like
        list of sample times

    Raises
    ------
    ValueError
        if values in ``sample_times`` are not monotonously increasing

    """
    if np.any(np.diff(sample_times) <= 0):
        raise ValueError("'sample_times' must be increasing!")


def _get_sample_times(
    data: pd.DataFrame,
    saliva_type: str,
    sample_times: Optional[Union[np.array, Sequence[int]]] = None,
    remove_s0: Optional[bool] = False,
) -> np.array:
    if sample_times is None:
        # check if dataframe has 'time' index
        if "time" in data.index.names:
            data = data.reset_index("time")
        # check if dataframe has 'time' column
        if "time" in data.columns:
            sample_times = np.array(data.unstack(level="sample")["time"])
            if np.all(sample_times == sample_times[0]):
                # all subjects have the same saliva times
                sample_times = sample_times[0]
        else:
            raise ValueError("No sample times specified!")

    # ensure numpy
    sample_times = np.squeeze(sample_times)

    # check whether we have the same saliva times for all subjects (1d array) or not (2d array)
    # and whether the input format is correct
    sample_times = _sample_times_sanitize(data, sample_times, saliva_type)
    _get_sample_times_check_dims(data, sample_times, saliva_type)

    if remove_s0:
        sample_times = sample_times[..., 1:]
    return sample_times


def _sample_times_sanitize(data: pd.DataFrame, sample_times: arr_t, saliva_type: str) -> arr_t:
    if sample_times.ndim == 1:
        exp_shape = data.unstack(level="sample")[saliva_type].shape[1]
        act_shape = sample_times.shape[0]
        if act_shape != exp_shape and (act_shape % exp_shape) == 0:
            # saliva times are in long-format => number of sample times corresponds to 2nd dimension
            sample_times = np.array(sample_times.unstack("sample").squeeze())
    return sample_times


def _get_sample_times_check_dims(data: pd.DataFrame, sample_times: arr_t, saliva_type: str):
    if sample_times.ndim == 1:
        exp_shape = data.unstack(level="sample")[saliva_type].shape[1]
        act_shape = sample_times.shape[0]
        # saliva times equal for all subjects
        # => number of sample times must correspond to 2nd dimension of wide-format data
        if act_shape != exp_shape:
            raise ValueError(
                "'sample_times' does not correspond to the number of saliva samples in 'data'! "
                f"Expected {exp_shape}, got {act_shape}."
            )
    elif sample_times.ndim == 2:
        # saliva time different for all subjects
        exp_shape = data.unstack(level="sample")[saliva_type].shape
        act_shape = sample_times.shape
        if act_shape != exp_shape:
            raise ValueError(
                "Dimensions of 'sample_times' does not correspond to dimensions of 'data'! "
                f"Expected {exp_shape}, got {act_shape}."
            )
    else:
        raise ValueError(f"'sample_times' has invalid dimensions! Expected 1 or 2, got {sample_times.ndim}")


def _get_saliva_idx_labels(
    columns: pd.Index,
    sample_labels: Optional[Union[tuple, Sequence]] = None,
    sample_idx: Optional[Union[tuple[int, int], Sequence[int]]] = None,
) -> tuple[Sequence, Sequence]:
    """Get sample labels and indices from data, if only one of both was specified.

    Parameters
    ----------
    columns : :class:`pandas.Index`
        dataframe columns
    sample_labels : list, optional
        list of sample labels
    sample_idx : list, optional
        list of sample indices

    Returns
    -------
    sample_labels:
        list of sample labels
    sample_idx:
        list of sample indices

    """
    if sample_labels is not None:
        try:
            sample_idx = [columns.get_loc(label) for label in sample_labels]
        except KeyError as e:
            raise IndexError(f"Invalid sample_labels `{sample_labels}`") from e
    else:
        try:
            # ensure list
            sample_idx = list(sample_idx)
            sample_labels = columns[sample_idx]
        except IndexError as e:
            raise IndexError("`sample_idx[1]` is out of bounds!") from e

    if len(sample_idx) != 2:
        raise IndexError(f"Exactly 2 indices needed for computing slope. Got {len(sample_idx)} indices.")

    sample_idx = _get_saliva_idx_labels_sanitize(sample_idx, columns)
    return sample_labels, sample_idx


def _get_saliva_idx_labels_sanitize(sample_idx: list[int], columns: Sequence[str]):
    # replace idx values like '-1' with the actual index
    if sample_idx[0] < 0:
        sample_idx[0] = len(columns) + sample_idx[0]

    if sample_idx[1] < 0:
        sample_idx[1] = len(columns) + sample_idx[1]

    # check that second index is bigger than first index
    if sample_idx[0] >= sample_idx[1]:
        raise IndexError(f"`sample_idx[1]` must be bigger than `sample_idx[0]`. Got {sample_idx}")
    return sample_idx


def _get_group_cols(
    data: SalivaRawDataFrame, group_cols: Union[str, Sequence[str]], group_type: str, function_name: str
) -> list[str]:
    """Get appropriate columns for grouping.

    Parameters
    ----------
    data : :class:`~biopsykit.utils.dtypes.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    group_cols : str or list of str
        list of index levels and column names to group by
    group_type : str
        which index level should be grouped by: 'subject' (for computing features per subject, along 'sample' axis)
        or 'sample' (for computing values per sample, along 'subject' axis)
    function_name : str
        function name for error message: 'standard_features' or 'mean_se'

    Returns
    -------
    list of str
        list of group by columns

    """
    if isinstance(group_cols, str):
        # ensure list
        group_cols = [group_cols]
    elif group_cols is None:
        # group by all available index levels
        group_cols = list(data.index.names)
        group_cols.remove(group_type)

    if any(col not in list(data.index.names) + list(data.columns) for col in group_cols):
        # check for valid groupers
        raise ValueError(
            f"Computing {function_name} failed: Not all of '{group_cols}' are valid index levels or column names!"
        )
    return group_cols


_dict_saliva_type_suggs: dict[str, Sequence[str]] = {
    "cortisol": ["cortisol", "cort", "Cortisol", "_c_"],
    "amylase": ["amylase", "amy", "Amylase", "sAA"],
    "il6": ["il6", "IL6", "il-6", "IL-6", "il_6", "IL_6"],
}
"""Dictionary containing possible column patterns for different saliva types."""
